-- Paraselene's WildRace
-- Version 0.5.2

require "Apollo"
require "Window"
 
local WINDOW_VERTICAL_GAP = 3

ParaseleneWildRace.About = {}

function ParaseleneWildRace.About:new(app)
    local i = {}
    setmetatable(i, self)
    self.__index = self

    i.app = app

    -- Center the window above the default position of the client/satallite window group.
    -- See CharacterData.CreateDefaultClientWindowInfo() for window position details.
    local display = Apollo.GetDisplaySize()
    local left = (display and math.floor((display.nWidth - ParaseleneWildRace.minAboutWindowWidth) / 2)) or 0
    local top = (display and
                 (math.floor((display.nHeight - ParaseleneWildRace.minClientWindowHeight) / 2) -
                  WINDOW_VERTICAL_GAP -
                  ParaseleneWildRace.minAboutWindowHeight
                 )
                ) or 0

    i.wndAbout = Apollo.LoadForm(i.app.doc, "wndAbout", nil, i)
    if not i.wndAbout then
        local message = ParaseleneWildRace.appName ..
            ": Error: Could not load the \"About\" window."
        ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_System, message)
        Apollo.AddAddonErrorText(ParaseleneWildRaceInstance, message)
        return nil
    end

    local wndTitle = i.wndAbout:FindChild("wndTitle")

    i.edtDetails = i.wndAbout:FindChild("wndDetails"):FindChild("edtDetails")

    wndTitle:SetText("About " .. ParaseleneWildRace.appNameFull)
    
    i.wndAbout:SetSizingMinimum(ParaseleneWildRace.minAboutWindowWidth, ParaseleneWildRace.minAboutWindowHeight)
    i.wndAbout:Move(left, top, ParaseleneWildRace.minAboutWindowWidth, ParaseleneWildRace.minAboutWindowHeight)

    i.wndAbout:SetSprite(ParaseleneWildRace.backgroundSprites[i.app.characterData.uiOptions.background])
    i.wndAbout:Show(false, true)

    return i
end

function ParaseleneWildRace.About:HideWindow()
    if self.wndAbout:IsVisible() then
        self.wndAbout:Close()
    end
end

function ParaseleneWildRace.About:OnBtnClose(wndHandler, wndControl, eMouseButton)
    self:HideWindow()
    Sound.Play(ParaseleneWildRace.buttonClickSound)
end

function ParaseleneWildRace.About:SetContent(currentPosition)
    if self.wndAbout:IsVisible() then
        local contents = "Version " .. ParaseleneWildRace.appVersion.full .. "\n" ..
            "Written by Paraselene Alqwi\n\n"

        contents = contents .. "Current location:\n"
        if currentPosition then
            contents = contents ..
                currentPosition.subZone .. "\n" .. currentPosition.zone .. "\n" .. currentPosition.realm .. "\n" ..
                string.format("%.0f, %.0f, %.0f",
                    currentPosition.point.x, currentPosition.point.z, currentPosition.point.y
                ) .. "\n" ..
                "Accessible by " .. (
                    ((currentPosition.faction == ParaseleneWildRace.waypointFaction.all) and "everyone") or
                    ((currentPosition.faction == ParaseleneWildRace.waypointFaction.dominion) and "Dominion only") or
                    ((currentPosition.faction == ParaseleneWildRace.waypointFaction.exile) and "Exile only")
                ) .. "\n\n"
        else
            contents = contents .. "Unknown\n\n"
        end

        contents = contents .. "Slash command: /wildrace\n\n" ..

            "Paraselene's sisters (alts) are:\n" ..
            "Parahelia Alqwi\n" ..
            "Elyzabeth Alqwi\n" ..
            "Abigail Alqwi\n\n" ..

            "Paraselene's coven-mates (alts) are:\n" ..
            "Irina Feytal\n" ..
            "Wilma Twygdance\n" ..
            "Veronyca Nytesharp\n" ..
            "Jeka Sandstone\n" ..
            "Closure Icosaciph\n" ..
            "Ada Octapatch\n" ..
            "Megrim Zir\n\n" ..

            "Thank you to those in my guilds,\n" ..
            "'StarLords' and 'Story of Crows', for\n" ..
            "aiding me in testing.\n\n" ..

            "A very special thanks to Katia Managan,\n" ..
            "who's interactive add-ons inspired\n" ..
            "me to create WildRace."
            
        self.edtDetails:SetText(contents)
    end
end

function ParaseleneWildRace.About:ShowWindow()
    if not self.wndAbout:IsVisible() then
        self.wndAbout:Invoke()
    end
end

function ParaseleneWildRace.About:ToggleWindowVisibility()
    if self.wndAbout:IsVisible() then
        self:HideWindow()
    else
        self:ShowWindow()
    end
end
