-- Paraselene's WildRace
-- Version 0.5.2

local ACCOUNT_DATA_FORMAT_VERSION = 4

ParaseleneWildRace.AccountData = {}

function ParaseleneWildRace.AccountData:new(o)
    if type(o) ~= "table" then
        return nil
    end

    local i = {}
    setmetatable(i, self)
    self.__index = self 

    i.version = ((type(o.version) == "number") and o.version) or nil
    local restoreSucceeded = false
    if (i.version == 1) or (i.version == 2) then
        restoreSucceeded = i:RestoreV1V2(o)
    elseif i.version == 3 then
        restoreSucceeded = i:RestoreV3(o)
    elseif i.version == 4 then
        restoreSucceeded = i:RestoreV4(o)
    end
    i.version = ACCOUNT_DATA_FORMAT_VERSION

    return (restoreSucceeded and i) or nil
end

function ParaseleneWildRace.AccountData.CreateDefault()
    return ParaseleneWildRace.AccountData:new{
        version = ACCOUNT_DATA_FORMAT_VERSION,
        storage = ParaseleneWildRace.AccountData.CreateDefaultStorage()
    }
end

function ParaseleneWildRace.AccountData.CreateDefaultStorage()
    local baseFolder = ParaseleneWildRace.Folder:new{
        name = "Base"
    }
    local paraseleneFolder = ParaseleneWildRace.Folder:new{
        name = "Paraselene's Favorites"
    }
    local zoneFolder = ParaseleneWildRace.Folder:new{
        name = "Zone Courses"
    }
    local illiumFolder = ParaseleneWildRace.Folder:new{
        name = "Illium Courses"
    }
    local thaydFolder = ParaseleneWildRace.Folder:new{
        name = "Thayd Courses"
    }
    local whitevaleFolder = ParaseleneWildRace.Folder:new{
        name = "Whitevale Courses"
    }

    local paraseleneDominionCourse = ParaseleneWildRace.Course:new{
        name = "Paraselene's Dominion coven-mates (alts)",
        strictMode = false,
        timedMode = false,
        timeout = 0
    }
    paraseleneDominionCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "House Alqwi Manor",
        realm = "Entity",
        zone = "Parahelia Alqwi",
        subZone = "House Alqwi Manor",
        inHousing = true,
        faction = ParaseleneWildRace.waypointFaction.dominion,
        point = ParaseleneWildRace.Point:new{x = 1574, y = -708, z = 1323}
    })
    paraseleneDominionCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Alqwi Family Farm",
        realm = "Entity",
        zone = "Elyzabeth Alqwi",
        subZone = "Alqwi Family Farm",
        inHousing = true,
        faction = ParaseleneWildRace.waypointFaction.dominion,
        point = ParaseleneWildRace.Point:new{x = 1574, y = -708, z = 1323}
    })
    paraseleneDominionCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Beneath",
        realm = "Entity",
        zone = "Closure Icosaciph",
        subZone = "Beneath",
        inHousing = true,
        faction = ParaseleneWildRace.waypointFaction.dominion,
        point = ParaseleneWildRace.Point:new{x = 1574, y = -708, z = 1323}
    })
    paraseleneDominionCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Flood Site X19",
        realm = "Entity",
        zone = "Ada Octapatch",
        subZone = "Flood Site X19",
        inHousing = true,
        faction = ParaseleneWildRace.waypointFaction.dominion,
        point = ParaseleneWildRace.Point:new{x = 1574, y = -708, z = 1323}
    })
    paraseleneDominionCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Megrim Crater",
        realm = "Entity",
        zone = "Megrim Zir",
        subZone = "Megrim Crater",
        inHousing = true,
        faction = ParaseleneWildRace.waypointFaction.dominion,
        point = ParaseleneWildRace.Point:new{x = 1574, y = -708, z = 1323}
    })

    local paraseleneExileCourse = ParaseleneWildRace.Course:new{
        name = "Paraselene and her Exile coven-mates (alts)",
        strictMode = false,
        timedMode = false,
        timeout = 0
    }
    paraseleneExileCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Alqwi Aerodrome",
        realm = "Entity",
        zone = "Paraselene Alqwi",
        subZone = "Alqwi Aerodrome",
        inHousing = true,
        faction = ParaseleneWildRace.waypointFaction.exile,
        point = ParaseleneWildRace.Point:new{x = 1574, y = -708, z = 1323}
    })
    paraseleneExileCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Feytal Research Hospital",
        realm = "Entity",
        zone = "Irina Feytal",
        subZone = "Feytal Research Hospital",
        inHousing = true,
        faction = ParaseleneWildRace.waypointFaction.exile,
        point = ParaseleneWildRace.Point:new{x = 1574, y = -708, z = 1323}
    })
    paraseleneExileCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Rainwater Grove",
        realm = "Entity",
        zone = "Wilma Twygdance",
        subZone = "Rainwater Grove",
        inHousing = true,
        faction = ParaseleneWildRace.waypointFaction.exile,
        point = ParaseleneWildRace.Point:new{x = 1574, y = -708, z = 1323}
    })
    paraseleneExileCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Chillwind Spire",
        realm = "Entity",
        zone = "Veronyca Nytesharp",
        subZone = "Chillwind Spire",
        inHousing = true,
        faction = ParaseleneWildRace.waypointFaction.exile,
        point = ParaseleneWildRace.Point:new{x = 1574, y = -708, z = 1323}
    })
    paraseleneExileCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Blue Sky Mine",
        realm = "Entity",
        zone = "Jeka Sandstone",
        subZone = "Blue Sky Mine",
        inHousing = true,
        faction = ParaseleneWildRace.waypointFaction.exile,
        point = ParaseleneWildRace.Point:new{x = 1574, y = -708, z = 1323}
    })
    paraseleneExileCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Abigail Rock",
        realm = "Entity",
        zone = "Abigail Alqwi",
        subZone = "Abigail Rock",
        inHousing = true,
        faction = ParaseleneWildRace.waypointFaction.exile,
        point = ParaseleneWildRace.Point:new{x = 1574, y = -708, z = 1323}
    })

    local illiumCourse = ParaseleneWildRace.Course:new{
        name = "Around Illium",
        strictMode = true,
        timedMode = true,
        timeout = 240
    }
    illiumCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Spaceport Alpha",
        realm = "Entity",
        zone = "Illium",
        subZone = "Spaceport Alpha",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = -3398, y = -891, z = -702}
    })
    illiumCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Parade Grounds",
        realm = "Entity",
        zone = "Illium",
        subZone = "Parade Grounds",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = -3477, y = -885, z = -988}
    })
    illiumCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Fate's Landing",
        realm = "Entity",
        zone = "Illium",
        subZone = "Fate's Landing",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = -3123, y = -889, z = -1179}
    })
    illiumCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Legion's Way",
        realm = "Entity",
        zone = "Illium",
        subZone = "Legion's Way",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = -2914, y = -896, z = -706}
    })
    illiumCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "South East Gate",
        realm = "Entity",
        zone = "Illium",
        subZone = "Legion's Way",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = -2900, y = -905, z = -448}
    })
    illiumCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Luminous Gardens",
        realm = "Entity",
        zone = "Illium",
        subZone = "Luminous Gardens",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = -3050, y = -908, z = -694}
    })
    illiumCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Protostar Housing Initiative",
        realm = "Entity",
        zone = "Illium",
        subZone = "Protostar Housing Initiative",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = -3246, y = -905, z = -821}
    })
    illiumCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Conqueror's Square",
        realm = "Entity",
        zone = "Illium",
        subZone = "Conqueror's Square",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = -3320, y = -887, z = -529}
    })
    illiumCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "West Gate",
        realm = "Entity",
        zone = "Illium",
        subZone = "Spaceport Alpha",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = -3818, y = -865, z = -704}
    })
    illiumCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Spaceport Alpha",
        realm = "Entity",
        zone = "Illium",
        subZone = "Spaceport Alpha",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = -3398, y = -891, z = -702}
    })

    local thaydCourse = ParaseleneWildRace.Course:new{
        name = "Around Thayd",
        strictMode = true,
        timedMode = true,
        timeout = 180
    }
    thaydCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Spaceport Horizon",
        realm = "Entity",
        zone = "Thayd",
        subZone = "Spaceport Horizon",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = 4082, y = -803, z = -2354}
    })
    thaydCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Academy Corner",
        realm = "Entity",
        zone = "Thayd",
        subZone = "Academy Corner",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = 4237, y = -815, z = -2408}
    })
    thaydCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "FCON Headquarters",
        realm = "Entity",
        zone = "Thayd",
        subZone = "FCON Headquarters",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = 4234, y = -811, z = -2201}
    })
    thaydCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Alchemist's Maze",
        realm = "Entity",
        zone = "Thayd",
        subZone = "Alchemist's Maze",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = 4403, y = -787, z = -1890}
    })
    thaydCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Walker's Bluff",
        realm = "Entity",
        zone = "Thayd",
        subZone = "Walker's Bluff",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = 4234, y = -714, z = -1707}
    })
    thaydCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Protostar Housing Initiative",
        realm = "Entity",
        zone = "Thayd",
        subZone = "Protostar Housing Initiative",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = 4033, y = -822, z = -1699}
    })
    thaydCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Shindig Hollow",
        realm = "Entity",
        zone = "Thayd",
        subZone = "Shindig Hollow",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = 3710, y = -843, z = -1698}
    })
    thaydCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Arborian Gardens",
        realm = "Entity",
        zone = "Thayd",
        subZone = "Arborian Gardens",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = 3703, y = -845, z = -1962}
    })
    thaydCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Fortune's Ground",
        realm = "Entity",
        zone = "Thayd",
        subZone = "Fortune's Ground",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = 3926, y = -843, z = -1898}
    })
    thaydCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Court of Judgment",
        realm = "Entity",
        zone = "Thayd",
        subZone = "Court of Judgment",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = 3948, y = -811, z = -2049}
    })
    thaydCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Traverse Tunnels",
        realm = "Entity",
        zone = "Thayd",
        subZone = "Traverse Tunnels",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = 3968, y = -829, z = -2183}
    })
    thaydCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Spaceport Horizon",
        realm = "Entity",
        zone = "Thayd",
        subZone = "Spaceport Horizon",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = 4082, y = -803, z = -2354}
    })

    local whitevaleCourse = ParaseleneWildRace.Course:new{
        name = "Central Whitevale",
        strictMode = true,
        timedMode = true,
        timeout = 420
    }
    whitevaleCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Krag and Kezzia",
        realm = "Entity",
        zone = "Whitevale",
        subZone = "Wigwalli Village",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = 3469, y = -940, z = -409}
    })
    whitevaleCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Platform 1",
        realm = "Entity",
        zone = "Whitevale",
        subZone = "Redmoon's Dockyard",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = 3724, y = -916, z = -175}
    })
    whitevaleCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Platform 2",
        realm = "Entity",
        zone = "Whitevale",
        subZone = "Redmoon's Dockyard",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = 3927, y = -901, z = -134}
    })
    whitevaleCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Upper Rimewater River",
        realm = "Entity",
        zone = "Whitevale",
        subZone = "Rimewater River",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = 3347, y = -957, z = -51}
    })
    whitevaleCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Inception Bridge",
        realm = "Entity",
        zone = "Whitevale",
        subZone = "Inception's Hold",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = 3186, y = -956, z = 149}
    })
    whitevaleCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Rimewater Falls",
        realm = "Entity",
        zone = "Whitevale",
        subZone = "Rimewater River",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = 3269, y = -967, z = 598}
    })
    whitevaleCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Lower Rimewater River",
        realm = "Entity",
        zone = "Whitevale",
        subZone = "South Whitevale",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = 3281, y = -1009, z = 697}
    })
    whitevaleCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "South Whitevale",
        realm = "Entity",
        zone = "Whitevale",
        subZone = "South Whitevale",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = 3650, y = -955, z = 753}
    })
    whitevaleCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Outbreak Fen",
        realm = "Entity",
        zone = "Whitevale",
        subZone = "Outbreak Fen",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = 3936, y = -960, z = 409}
    })
    whitevaleCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Genesis Installation",
        realm = "Entity",
        zone = "Whitevale",
        subZone = "Genesis Installation",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = 4107, y = -945, z = 184}
    })
    whitevaleCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "West Shore",
        realm = "Entity",
        zone = "Whitevale",
        subZone = "Thermock Lake West",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = 4075, y = -952, z = -29}
    })
    whitevaleCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Underwater Hatch",
        realm = "Entity",
        zone = "Whitevale",
        subZone = "Fortune's Fool",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = 4145, y = -997, z = -193}
    })
    whitevaleCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Mouth of the Cave",
        realm = "Entity",
        zone = "Whitevale",
        subZone = "Darkice Cave",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = 3916, y = -907, z = -1070}
    })
    whitevaleCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Wigwalli Festival Grounds",
        realm = "Entity",
        zone = "Whitevale",
        subZone = "Wigwalli Festival Grounds",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = 3580, y = -938, z = -526}
    })
    whitevaleCourse:Append(ParaseleneWildRace.Waypoint:new{
        name = "Krag and Kezzia",
        realm = "Entity",
        zone = "Whitevale",
        subZone = "Wigwalli Village",
        inHousing = false,
        faction = ParaseleneWildRace.waypointFaction.all,
        point = ParaseleneWildRace.Point:new{x = 3469, y = -940, z = -409}
    })

    paraseleneFolder:Append(paraseleneDominionCourse)
    paraseleneFolder:Append(paraseleneExileCourse)
    illiumFolder:Append(illiumCourse)
    thaydFolder:Append(thaydCourse)
    whitevaleFolder:Append(whitevaleCourse)
    zoneFolder:Append(illiumFolder)
    zoneFolder:Append(thaydFolder)
    zoneFolder:Append(whitevaleFolder)
    baseFolder:Append(paraseleneFolder)
    baseFolder:Append(zoneFolder)

    return ParaseleneWildRace.Storage:new{
        baseFolder = baseFolder
    }
end

function ParaseleneWildRace.AccountData:IsValid()
    -- self.version is not checked here.
    return (self.storage ~= nil)
end

function ParaseleneWildRace.AccountData:RestoreV1V2(o)
    -- self.version is not restored here.
    -- Note: V1 and V2 settings are not preserved.
    self.storage = self.CreateDefaultStorage()

    return self:IsValid()
end

function ParaseleneWildRace.AccountData:RestoreV3(o)
    -- self.version is not restored here.
    self.storage = ParaseleneWildRace.Storage:new(o.storage)

    -- Save the current course into storage.
    local currentCourse = ParaseleneWildRace.Course:new(o.currentCourse)
    if currentCourse and self.storage and self.storage.baseFolder then
        currentCourse.name = "Pre-version-0.5.0 current course: " .. currentCourse.name
        self.storage.baseFolder:Append(currentCourse)
    end

    return self:IsValid()
end

function ParaseleneWildRace.AccountData:RestoreV4(o)
    -- self.version is not restored here.
    self.storage = ParaseleneWildRace.Storage:new(o.storage)

    return self:IsValid()
end

function ParaseleneWildRace.AccountData:Save()
    return {
        version = ACCOUNT_DATA_FORMAT_VERSION,
        storage = self.storage:Save()
    }
end
