-- Paraselene's WildRace
-- Version 0.5.2

require "Apollo"

local ENABLE_TIMER_TICK_INTERVAL_SECS = 0.2
local ENABLE_TIMER_EXPIRE_SECS = 2.0

ParaseleneWildRace.Button = {}

function ParaseleneWildRace.Button:new(o)
    if type(o) ~= "table" then
        return nil
    end

    local i = {}
    setmetatable(i, self)
    self.__index = self

    i.button = o.button

    i.style = o.style or ParaseleneWildRace.buttonStyles.normal

    if i.style == ParaseleneWildRace.buttonStyles.red then
        i.disabledBorder = ParaseleneWildRace.disabledRedButtonBorder
        i.enabledBorder = ParaseleneWildRace.enabledRedButtonBorder
        i.flybyBorder = ParaseleneWildRace.enabledRedButtonFlybyBorder
    else -- if i.style = ParaseleneWildRace.buttonStyles.normal then
        i.disabledBorder = o.disabledBorder or ParaseleneWildRace.disabledButtonBorder
        i.enabledBorder = o.enabledBorder or ParaseleneWildRace.enabledButtonBorder
        i.flybyBorder = o.flybyBorder or ParaseleneWildRace.enabledButtonFlybyBorder
    end

    i.enabled = o.enabled

    i.icon = i.button:FindChild("wndIcon")
    i.progressBar = i.icon:FindChild("progressBar")
    if i.progressBar then
        i:ResetProgressBar()
        i.ticksMax = ENABLE_TIMER_EXPIRE_SECS / ENABLE_TIMER_TICK_INTERVAL_SECS
        i.progressBar:SetMax(i.ticksMax)
    end

    return i
end

function ParaseleneWildRace.Button:Disable()
    self.icon:SetTextColor(ApolloColor.new(ParaseleneWildRace.disabledTextColor))
    self.icon:SetSprite(self.disabledBorder)
    if self.progressBar then
        self.progressBar:SetTextColor(ApolloColor.new(ParaseleneWildRace.disabledTextColor))
        self.progressBar:SetSprite(self.disabledBorder)
        if self.enableTimer then
            self.enableTimer:Stop()
            self.enableTimer = nil
        end
        self:ResetProgressBar()
    end
    self.enabled = false
    self.button:Enable(self.enabled)
end

function ParaseleneWildRace.Button:Enable()
    self.icon:SetTextColor(ApolloColor.new(ParaseleneWildRace.enabledTextColor))
    if self.progressBar then
        self.icon:SetSprite(self.enabledBorder)
        self.progressBar:SetTextColor(ApolloColor.new(ParaseleneWildRace.enabledTextColor))
        self.progressBar:SetSprite(self.enabledBorder)
        if self.button:ContainsMouse() then
            self:ResetProgressBar()
            if not self.enableTimer then
                self.enableTimer = ApolloTimer.Create(ENABLE_TIMER_TICK_INTERVAL_SECS, true, "OnEnableTimer", self)
            end
        end
        self:ResetProgressBar()
    else
        self.icon:SetSprite((self.button:ContainsMouse() and self.flybyBorder) or self.enabledBorder)
    end
    self.enabled = true
    self.button:Enable(self.enabled)
end

function ParaseleneWildRace.Button:IsDisabled()
    return not self:IsEnabled()
end

function ParaseleneWildRace.Button:IsEnabled()
    if self.progressBar then
        return self.enabled and (self.ticks == self.ticksMax)
    end
    return self.enabled
end

function ParaseleneWildRace.Button:OnEnableTimer()
    if self.ticks < self.ticksMax then
        self.ticks = self.ticks + 1
        self.progressBar:SetProgress(self.ticks)
    else
        self.enableTimer:Stop()
        self.enableTimer = nil
    end
end

function ParaseleneWildRace.Button:OnMouseEnter()
    if self.enabled then
        if self.progressBar then
            self.icon:SetSprite(self.enabledBorder)
            self.progressBar:SetSprite(self.enabledBorder)
            self:ResetProgressBar()
            if not self.enableTimer then
                self.enableTimer = ApolloTimer.Create(ENABLE_TIMER_TICK_INTERVAL_SECS, true, "OnEnableTimer", self)
            end
        else
            self.icon:SetSprite(self.flybyBorder)
        end
    end
end

function ParaseleneWildRace.Button:OnMouseExit()
    if self.enabled then
        self.icon:SetSprite(self.enabledBorder)
        if self.progressBar then
            self.progressBar:SetSprite(self.enabledBorder)
            if self.enableTimer then
                self.enableTimer:Stop()
                self.enableTimer = nil
            end
            self:ResetProgressBar()
        end
    end
end

function ParaseleneWildRace.Button:ResetProgressBar()
    self.ticks = 0
    self.progressBar:SetProgress(self.ticks)
end
