-- Paraselene's WildRace
-- Version 0.5.2

require "Apollo"
require "CombatFloater"
require "HousingLib"
require "Window"
 
local RACE_POSITION_TIMER_TICK_INTERVAL_SECS = 0.25
local RACE_CLOCK_TIMER_TICK_INTERVAL_SECS = 1.0

ParaseleneWildRace.Client = {}

function ParaseleneWildRace.Client:new(app)
    local i = {}
    setmetatable(i, self)
    self.__index = self

    i.app = app

    if not i.app.characterData.clientWindowInfo.width or
       (i.app.characterData.clientWindowInfo.width < ParaseleneWildRace.minClientWindowWidth) then
        i.app.characterData.clientWindowInfo.width = ParaseleneWildRace.minClientWindowWidth
    end
    if not i.app.characterData.clientWindowInfo.height or
       (i.app.characterData.clientWindowInfo.height < ParaseleneWildRace.minClientWindowHeight) then
        i.app.characterData.clientWindowInfo.height = ParaseleneWildRace.minClientWindowHeight
    end

    i.wndClient = Apollo.LoadForm(i.app.doc, "wndClient", nil, i)
    if not i.wndClient then
        local message = ParaseleneWildRace.appName ..
            ": Error: Could not load the client window."
        ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_System, message)
        Apollo.AddAddonErrorText(ParaseleneWildRaceInstance, message)
        return nil
    end

    local wndTitle = i.wndClient:FindChild("wndTitle")
    i.btnAbout = wndTitle:FindChild("btnAbout")
    i.btnLock = wndTitle:FindChild("btnLock")
    i.wndLockIcon = i.btnLock:FindChild("wndIcon")

    local wndToolbar = i.wndClient:FindChild("wndToolbar")
    i.btnStartStop = wndToolbar:FindChild("btnStartStop")
    i.wndStartStopFrame = i.btnStartStop:FindChild("wndFrame")
    i.wndStartStopIcon = i.wndStartStopFrame:FindChild("wndIcon")
    i.wndStartStopDisabled = i.wndStartStopIcon:FindChild("wndDisabled")
    i.btnVisitHome = wndToolbar:FindChild("btnVisitHome")
    i.wndVisitHomeFrame = i.btnVisitHome:FindChild("wndFrame")
    i.wndVisitHomeDisabled = i.wndVisitHomeFrame:FindChild("wndIcon"):FindChild("wndDisabled")
    i.btnVisibleWaypoints = wndToolbar:FindChild("btnVisibleWaypoints")
    i.wndVisibleWaypointsFrame = i.btnVisibleWaypoints:FindChild("wndFrame")
    i.btnOptions = wndToolbar:FindChild("btnOptions")
    i.wndOptionsFrame = i.btnOptions:FindChild("wndFrame")
    i.btnEditor = wndToolbar:FindChild("btnEditor")
    i.wndEditorFrame = i.btnEditor:FindChild("wndFrame")
    i.wndEditorDisabled = i.wndEditorFrame:FindChild("wndIcon"):FindChild("wndDisabled")
    i.btnFiles = wndToolbar:FindChild("btnFiles")
    i.wndFilesFrame = i.btnFiles:FindChild("wndFrame")
    i.wndFilesDisabled = i.wndFilesFrame:FindChild("wndIcon"):FindChild("wndDisabled")
    i.btnShare = wndToolbar:FindChild("btnShare")
    i.wndShareFrame = i.btnShare:FindChild("wndFrame")
    i.wndShareDisabled = i.wndShareFrame:FindChild("wndIcon"):FindChild("wndDisabled")

    local wndCourse = i.wndClient:FindChild("wndCourse")
    i.edtCourseName = wndCourse:FindChild("edtName")
    i.wndWaypointList = wndCourse:FindChild("wndWaypointList")

    local wndWaypointDetails = i.wndClient:FindChild("wndWaypointDetails")
    i.edtWaypointName = wndWaypointDetails:FindChild("edtName")
    i.wndWaypointDetailsHeaders = wndWaypointDetails:FindChild("wndHeaders")
    i.edtWaypointDetails = wndWaypointDetails:FindChild("edtDetails")

    local wndStatusbar = i.wndClient:FindChild("wndStatusbar")
    i.wndStatus = wndStatusbar:FindChild("wndStatus")
    i.wndTimer = wndStatusbar:FindChild("wndTimer")

    wndTitle:SetText(ParaseleneWildRace.appNameFull)

    i.wndClient:SetSizingMinimum(ParaseleneWildRace.minClientWindowWidth, ParaseleneWildRace.minClientWindowHeight)
    i.wndClient:Move(
        i.app.characterData.clientWindowInfo.left,
        i.app.characterData.clientWindowInfo.top,
        i.app.characterData.clientWindowInfo.width,
        i.app.characterData.clientWindowInfo.height
    )
    i:SetWindowLock(i.app.characterData.clientWindowInfo.locked)

    i:SetCourse(nil)

    i.wndClient:SetSprite(ParaseleneWildRace.backgroundSprites[i.app.characterData.uiOptions.background])
    i.wndClient:Show(false, true)

    return i
end

function ParaseleneWildRace.Client:AdjustScrollbar()
    if self.index then
        local windowHeight = self.wndWaypointList:GetHeight()
        local listHeight = #self.course.waypoints * ParaseleneWildRace.waypointListItemHeight
        local waypointOffset = (self.index - 1) * ParaseleneWildRace.waypointListItemHeight
        local currentScrollPosition = self.wndWaypointList:GetVScrollPos()
        local newScrollPosition = currentScrollPosition
        if self.index > 2 then
            if newScrollPosition > (waypointOffset - ParaseleneWildRace.waypointListItemHeight) then
                newScrollPosition = waypointOffset - ParaseleneWildRace.waypointListItemHeight
            end
        else
            newScrollPosition = 0
        end
        if self.index < #self.course.waypoints then
            if (newScrollPosition + windowHeight) < (waypointOffset + (2 * ParaseleneWildRace.waypointListItemHeight)) then
                newScrollPosition = waypointOffset + (2 * ParaseleneWildRace.waypointListItemHeight) - windowHeight
            end
        else
            if (newScrollPosition + windowHeight) < (waypointOffset + ParaseleneWildRace.waypointListItemHeight) then
                newScrollPosition = waypointOffset + ParaseleneWildRace.waypointListItemHeight - windowHeight
            end
        end
        if newScrollPosition ~= currentScrollPosition then
            self.wndWaypointList:SetVScrollPos(newScrollPosition)
        end
    end
end

-- wndStatus and wndTimer are updated in OnRacePositionTimer() and OnUpdateTimer(),
-- so this function should not be called while either the updateTimer or racePositionTimer is running.
function ParaseleneWildRace.Client:AdjustStatusbar()
    -- Adjust the status message.
    if self.course then
        local status = ((self.course.timedMode and "Timed, ") or "Untimed, ")
        status = status .. ((self.course.strictMode and "strict") or "loose") .. " order - "
        self.wndStatus:SetText(status .. ((self.course.timedMode and "Remaining:") or "Elapsed:"))
    else
        self.wndStatus:SetText("")
    end

    -- Adjust the timer value.
    if self.course then
        self.wndTimer:SetText(self.FormatTime((self.course.timedMode and self.course.timeout) or 0))
    else
        self.wndTimer:SetText("")
    end
end

function ParaseleneWildRace.Client:AdjustToolbar()
    -- Adjust the start/stop button.
    if self.isRaceInProgress then
        self.isStartStopEnabled = true
        self.wndStartStopFrame:SetSprite(
            (self.btnStartStop:ContainsMouse() and ParaseleneWildRace.pressedToolbarButtonFlybyFrame) or
            ParaseleneWildRace.pressedToolbarButtonFrame
        )
        self.wndStartStopIcon:SetSprite(ParaseleneWildRace.stopIcon)
        self.wndStartStopDisabled:SetSprite("")
    else
        self.isStartStopEnabled = self:GetFirstVisitableIndex() ~= nil
        if self.isStartStopEnabled then
            self.wndStartStopFrame:SetSprite(
                (self.btnStartStop:ContainsMouse() and ParaseleneWildRace.enabledToolbarButtonFlybyFrame) or
                ParaseleneWildRace.enabledToolbarButtonFrame
            )
        else
            self.wndStartStopFrame:SetSprite(ParaseleneWildRace.disabledToolbarButtonFrame)
        end
        self.wndStartStopIcon:SetSprite(ParaseleneWildRace.startIcon)
        self.wndStartStopDisabled:SetSprite((self.isStartStopEnabled and "") or ParaseleneWildRace.noIcon)
    end
    self.btnStartStop:Enable(self.isStartStopEnabled)

    -- The visit home button is handled by OnUpdateTimer().

    -- Adjust the visible waypoints button.
    if self.app.characterData.areWaypointsVisible then
        self.wndVisibleWaypointsFrame:SetSprite(
            (self.btnVisibleWaypoints:ContainsMouse() and ParaseleneWildRace.pressedToolbarButtonFlybyFrame) or
            ParaseleneWildRace.pressedToolbarButtonFrame
        )
    else
        self.wndVisibleWaypointsFrame:SetSprite(
            (self.btnVisibleWaypoints:ContainsMouse() and ParaseleneWildRace.enabledToolbarButtonFlybyFrame) or
            ParaseleneWildRace.enabledToolbarButtonFrame
        )
    end

    -- Adjust the Options button.
    if self.btnOptions:ContainsMouse() then
        self.wndOptionsFrame:SetSprite(
            (self.app.options.wndOptions:IsVisible() and ParaseleneWildRace.pressedToolbarButtonFlybyFrame) or
            ParaseleneWildRace.enabledToolbarButtonFlybyFrame
        )
    else
        self.wndOptionsFrame:SetSprite(ParaseleneWildRace.enabledToolbarButtonFrame)
    end

    -- Adjust the Editor button.
    self.isEditorEnabled = not self.isRaceInProgress
    if self.isEditorEnabled then
        if self.btnEditor:ContainsMouse() then
            self.wndEditorFrame:SetSprite(
                (self.app.editor.wndEditor:IsVisible() and ParaseleneWildRace.pressedToolbarButtonFlybyFrame) or
                ParaseleneWildRace.enabledToolbarButtonFlybyFrame
            )
        else
            self.wndEditorFrame:SetSprite(ParaseleneWildRace.enabledToolbarButtonFrame)
        end
        self.wndEditorDisabled:SetSprite("")
    else
        self.wndEditorFrame:SetSprite(ParaseleneWildRace.disabledToolbarButtonFrame)
        self.wndEditorDisabled:SetSprite(ParaseleneWildRace.noIcon)
    end
    self.btnEditor:Enable(self.isEditorEnabled)

    -- Adjust the Files button.
    self.isFilesEnabled = not self.isRaceInProgress
    if self.isFilesEnabled then
        if self.btnFiles:ContainsMouse() then
            self.wndFilesFrame:SetSprite(
                (self.app.files.wndFiles:IsVisible() and ParaseleneWildRace.pressedToolbarButtonFlybyFrame) or
                ParaseleneWildRace.enabledToolbarButtonFlybyFrame
            )
        else
            self.wndFilesFrame:SetSprite(ParaseleneWildRace.enabledToolbarButtonFrame)
        end
        self.wndFilesDisabled:SetSprite("")
    else
        self.wndFilesFrame:SetSprite(ParaseleneWildRace.disabledToolbarButtonFrame)
        self.wndFilesDisabled:SetSprite(ParaseleneWildRace.noIcon)
    end
    self.btnFiles:Enable(self.isFilesEnabled)

    -- Adjust the Share button.
    self.isShareEnabled = self.course and not self.isRaceInProgress
    if self.isShareEnabled then
        if self.btnShare:ContainsMouse() then
            self.wndShareFrame:SetSprite(
                (self.app.share.wndShare:IsVisible() and ParaseleneWildRace.pressedToolbarButtonFlybyFrame) or
                ParaseleneWildRace.enabledToolbarButtonFlybyFrame
            )
        else
            self.wndShareFrame:SetSprite(ParaseleneWildRace.enabledToolbarButtonFrame)
        end
        self.wndShareDisabled:SetSprite("")
    else
        self.wndShareFrame:SetSprite(ParaseleneWildRace.disabledToolbarButtonFrame)
        self.wndShareDisabled:SetSprite(ParaseleneWildRace.noIcon)
    end
    self.btnShare:Enable(self.isShareEnabled)
end

function ParaseleneWildRace.Client.Broadcast(message, color)
    local options = {
        nColor = color,
        fDuration = 2,
        fFadeInDuration = 0,
        fEndHoldDuration = 1,
        fFadeOutDuration = 1,

        strFontFace = "CRB_FloaterLarge",
        fSpinAroundRadius = 0,
        fVibrate = 0,

        eCollisionMode = CombatFloater.CodeEnumFloaterCollisionMode.Vertical,
        fExpandCollisionBoxWidth = 1,
        fExpandCollisionBoxHeight = 1,
        eLocation = CombatFloater.CodeEnumFloaterLocation.Bottom,

        fOffsetDirection = 0,
        fOffset = 0,
        nDigitSpriteSpacing = 0,
        bUseScreenPos = true,
        bShowOnTop = true,
        fDelay = 0,

        fScale = 1,
        fRotation = 0,
        fVelocityDirection = 0,
        fVelocityMagnitude = 0,
        fAccelDirection = 0,
        fAccelMagnitude = 0,

        fExpand = 1
    }
    CombatFloater.ShowTextFloater(GameLib.GetControlledUnit(), message, options)
end

function ParaseleneWildRace.Client:CheckPosition(currentPosition, raceCourseWaypoint)
    if currentPosition and
       raceCourseWaypoint.isVisitable and
       not raceCourseWaypoint.hasBeenVisited and
       (raceCourseWaypoint.zone == currentPosition.zone) and
       (raceCourseWaypoint.inHousing == currentPosition.inHousing) and
       raceCourseWaypoint.cube:InRange(currentPosition.point) then

        raceCourseWaypoint.hasBeenVisited = true

         -- Set the checkmark on the waypoint in the list.
        raceCourseWaypoint.wndStatus:SetSprite(ParaseleneWildRace.checkmarkIcon)
        raceCourseWaypoint.wndStatus:SetTooltip("The waypoint has been visited")

        local message = "Waypoint \"" .. raceCourseWaypoint.name .. "\" reached"
        ChatSystemLib.PostOnChannel(
            ChatSystemLib.ChatChannel_System,
            ParaseleneWildRace.appName .. ": " .. message .. "."
        )
        self.Broadcast(message, ParaseleneWildRace.broadcastBlue)

        -- Advance the waypoint selection to the next un-visited, visitable waypoint in the list.
        local firstVisitableIndex = self:GetFirstVisitableIndex{skipVisited = true}
        if firstVisitableIndex then
            self:SetIndex(firstVisitableIndex)
            self:AdjustScrollbar()
        else
            -- No more un-visited, visitable waypoints. We have finished the race and won.
            self.isRaceFinished = true

            self.wndStatus:SetText("Finished in:")

            message = "Race finished"
            ChatSystemLib.PostOnChannel(
                ChatSystemLib.ChatChannel_System,
                ParaseleneWildRace.appName .. ": " .. message .. "."
            )
            self.Broadcast(message, ParaseleneWildRace.broadcastBlue)
        end
    end
end

function ParaseleneWildRace.Client:ClearCourse()
    self:Stop()
    if self.course then
        self.edtCourseName:SetText("")
        for idx, waypoint in ipairs(self.course.waypoints) do
            if waypoint.wndWaypointListItem then
                waypoint.wndWaypointListItem:Destroy()
                waypoint.wndWaypointListItem = nil
            end
            if waypoint.wndStatus then
                waypoint.wndStatus = nil
            end
        end
        self.course = nil

        -- index is set to nil here in order to maintain integrity between course and index.
        -- SetIndex() expects self.index to be nil if self.course.waypoints is empty.
        self.index = nil
    end

    -- Possible Carbine bug.
    -- Apparently, removing a scrolling window's children does not adjust its scrollbar; just adding them.
    -- Since our window will always be large than the space needed for a single waypoint, let's add a temporary child
    -- window to adjust the scrollbar, then delete the temporary child window.
    local wndTempWaypointListItem = Apollo.LoadForm(
        self.app.doc, "wndWaypointListItem", self.wndWaypointList, self
    )
    if not wndTempWaypointListItem then
        local message = ParaseleneWildRace.appName ..
            ": Error: Could not load a temporary waypoint list item window."
        ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_System, message)
        Apollo.AddAddonErrorText(ParaseleneWildRaceInstance, message)
    end
    wndTempWaypointListItem:Destroy()

    self:SetIndex(nil)

    self:AdjustToolbar()
    self:AdjustStatusbar()
    if self.app.editor then
        self.app.editor:AdjustRaceTools()
    end
    if self.app.files then
        self.app.files:AdjustTools()
    end
end

function ParaseleneWildRace.Client.FormatTime(seconds)
    return string.format(
        "%02.0f:%02.0f:%02.0f",
        math.floor(seconds / 3600),
        math.floor(seconds / 60) % 60,
        seconds % 60
    )
end

function ParaseleneWildRace.Client:GetFirstVisitableIndex(args)
    if self.course then
        for idx, waypoint in ipairs(self.course.waypoints) do
            if waypoint.isVisitable and not (waypoint.hasBeenVisited and args.skipVisited) then
                return idx
            end
        end
    end
    return nil
end

function ParaseleneWildRace.Client:GetScrollPosition()
    return self.wndWaypointList:GetVScrollPos()
end

function ParaseleneWildRace.Client:HideWindow()
    if self.wndClient:IsVisible() then

        self.app.about:HideWindow()

        -- When the client window closes, the satallite windows also close.
        -- When the client window re-opens, also re-open the satallite windows.
        self.wasEditorVisibleLastTimeClientWasVisible = self.app.editor.wndEditor:IsVisible()
        self.wasFilesVisibleLastTimeClientWasVisible = self.app.files.wndFiles:IsVisible()
        self.wasShareVisibleLastTimeClientWasVisible = self.app.share.wndShare:IsVisible()
        self.app.editor:HideWindow()
        self.app.files:HideWindow()
        self.app.options:HideWindow()
        self.app.share:HideWindow()

        self.wndClient:Close()
    end
end

function ParaseleneWildRace.Client:OnBtnAbout(wndHandler, wndControl, eMouseButton)
    self.app.about:ToggleWindowVisibility()
    Sound.Play(ParaseleneWildRace.buttonClickSound)
end

function ParaseleneWildRace.Client:OnBtnAboutGenerateTooltip(wndHandler, wndControl, eToolTipType, x, y)
    self.btnAbout:SetTooltip(
        (self.app.about.wndAbout:IsVisible() and "Close the \"About\" window") or
        "Open the \"About\" window"
    )
end

function ParaseleneWildRace.Client:OnBtnClose(wndHandler, wndControl, eMouseButton)
    self:HideWindow()
    Sound.Play(ParaseleneWildRace.buttonClickSound)
end

function ParaseleneWildRace.Client:OnBtnEditor(wndHandler, wndControl, eMouseButton)
    if self.isEditorEnabled then
        self.app.editor:ToggleWindowVisibility()
        if self.app.files.wndFiles:IsVisible() then
            self.app.files:HideWindow()
        end
        if self.app.options.wndOptions:IsVisible() then
            self.app.options:HideWindow()
        end
        if self.app.share.wndShare:IsVisible() then
            self.app.share:HideWindow()
        end
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    end
end

function ParaseleneWildRace.Client:OnBtnEditorGenerateTooltip(wndHandler, wndControl, eToolTipType, x, y)
    self.btnEditor:SetTooltip(
        (self.app.editor.wndEditor:IsVisible() and "Close the \"Editor\" window") or
        "Open the \"Editor\" window"
    )
end

function ParaseleneWildRace.Client:OnBtnEditorMouseEnter(wndHandler, wndControl, x, y)
    if self.isEditorEnabled then
        self.wndEditorFrame:SetSprite(
            (self.app.editor.wndEditor:IsVisible() and ParaseleneWildRace.pressedToolbarButtonFlybyFrame) or
            ParaseleneWildRace.enabledToolbarButtonFlybyFrame
        )
    else
        self.wndEditorFrame:SetSprite(ParaseleneWildRace.disabledToolbarButtonFrame)
    end
end

function ParaseleneWildRace.Client:OnBtnEditorMouseExit(wndHandler, wndControl, x, y)
    if self.isEditorEnabled then
        self.wndEditorFrame:SetSprite(
            (self.app.editor.wndEditor:IsVisible() and ParaseleneWildRace.pressedToolbarButtonFrame) or
            ParaseleneWildRace.enabledToolbarButtonFrame
        )
    else
        self.wndEditorFrame:SetSprite(ParaseleneWildRace.disabledToolbarButtonFrame)
    end
end

function ParaseleneWildRace.Client:OnBtnFiles(wndHandler, wndControl, eMouseButton)
    if self.isFilesEnabled then
        self.app.files:ToggleWindowVisibility()
        if self.app.editor.wndEditor:IsVisible() then
            self.app.editor:HideWindow()
        end
        if self.app.options.wndOptions:IsVisible() then
            self.app.options:HideWindow()
        end
        if self.app.share.wndShare:IsVisible() then
            self.app.share:HideWindow()
        end
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    end
end

function ParaseleneWildRace.Client:OnBtnFilesGenerateTooltip(wndHandler, wndControl, eToolTipType, x, y)
    self.btnFiles:SetTooltip(
        (self.app.files.wndFiles:IsVisible() and "Close the \"Files\" window") or
        "Open the \"Files\" window"
    )
end

function ParaseleneWildRace.Client:OnBtnFilesMouseEnter(wndHandler, wndControl, x, y)
    if self.isFilesEnabled then
        self.wndFilesFrame:SetSprite(
            (self.app.files.wndFiles:IsVisible() and ParaseleneWildRace.pressedToolbarButtonFlybyFrame) or
            ParaseleneWildRace.enabledToolbarButtonFlybyFrame
        )
    else
        self.wndFilesFrame:SetSprite(ParaseleneWildRace.disabledToolbarButtonFrame)
    end
end

function ParaseleneWildRace.Client:OnBtnFilesMouseExit(wndHandler, wndControl, x, y)
    if self.isFilesEnabled then
        self.wndFilesFrame:SetSprite(
            (self.app.files.wndFiles:IsVisible() and ParaseleneWildRace.pressedToolbarButtonFrame) or
            ParaseleneWildRace.enabledToolbarButtonFrame
        )
    else
        self.wndFilesFrame:SetSprite(ParaseleneWildRace.disabledToolbarButtonFrame)
    end
end

function ParaseleneWildRace.Client:OnBtnLock(wndHandler, wndControl, eMouseButton)
    self:SetWindowLock(not self.app.characterData.clientWindowInfo.locked)
    Sound.Play(ParaseleneWildRace.buttonClickSound)
end

function ParaseleneWildRace.Client:OnBtnLockGenerateTooltip(wndHandler, wndControl, eToolTipType, x, y)
    self.btnLock:SetTooltip(
        (self.app.characterData.clientWindowInfo.locked and "Unlock the window so that it can be moved and resized") or
        "Lock the window so that it cannot be moved or resized"
    )
end

function ParaseleneWildRace.Client:OnBtnOptions(wndHandler, wndControl, eMouseButton)
    self.app.options:ToggleWindowVisibility()
    if self.app.editor.wndEditor:IsVisible() then
        self.app.editor:HideWindow()
    end
    if self.app.files.wndFiles:IsVisible() then
        self.app.files:HideWindow()
    end
    if self.app.share.wndShare:IsVisible() then
        self.app.share:HideWindow()
    end
    Sound.Play(ParaseleneWildRace.buttonClickSound)
end

function ParaseleneWildRace.Client:OnBtnOptionsGenerateTooltip(wndHandler, wndControl, eToolTipType, x, y)
    self.btnOptions:SetTooltip(
        (self.app.options.wndOptions:IsVisible() and "Close the \"Options\" window") or
        "Open the \"Options\" window"
    )
end

function ParaseleneWildRace.Client:OnBtnOptionsMouseEnter(wndHandler, wndControl, x, y)
    self.wndOptionsFrame:SetSprite(
        (self.app.options.wndOptions:IsVisible() and ParaseleneWildRace.pressedToolbarButtonFlybyFrame) or
        ParaseleneWildRace.enabledToolbarButtonFlybyFrame
    )
end

function ParaseleneWildRace.Client:OnBtnOptionsMouseExit(wndHandler, wndControl, x, y)
    self.wndOptionsFrame:SetSprite(
        (self.app.options.wndOptions:IsVisible() and ParaseleneWildRace.pressedToolbarButtonFrame) or
        ParaseleneWildRace.enabledToolbarButtonFrame
    )
end

function ParaseleneWildRace.Client:OnBtnShare(wndHandler, wndControl, eMouseButton)
    if self.isShareEnabled then
        self.app.share:ToggleWindowVisibility()
        if self.app.editor.wndEditor:IsVisible() then
            self.app.editor:HideWindow()
        end
        if self.app.files.wndFiles:IsVisible() then
            self.app.files:HideWindow()
        end
        if self.app.options.wndOptions:IsVisible() then
            self.app.options:HideWindow()
        end
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    end
end

function ParaseleneWildRace.Client:OnBtnShareGenerateTooltip(wndHandler, wndControl, eToolTipType, x, y)
    self.btnShare:SetTooltip(
        (self.app.share.wndShare:IsVisible() and "Close the \"Share\" window") or
        "Open the \"Share\" window"
    )
end

function ParaseleneWildRace.Client:OnBtnShareMouseEnter(wndHandler, wndControl, x, y)
    if self.isShareEnabled then
        self.wndShareFrame:SetSprite(
            (self.app.share.wndShare:IsVisible() and ParaseleneWildRace.pressedToolbarButtonFlybyFrame) or
            ParaseleneWildRace.enabledToolbarButtonFlybyFrame
        )
    else
        self.wndShareFrame:SetSprite(ParaseleneWildRace.disabledToolbarButtonFrame)
    end
end

function ParaseleneWildRace.Client:OnBtnShareMouseExit(wndHandler, wndControl, x, y)
    if self.isShareEnabled then
        self.wndShareFrame:SetSprite(
            (self.app.share.wndShare:IsVisible() and ParaseleneWildRace.pressedToolbarButtonFrame) or
            ParaseleneWildRace.enabledToolbarButtonFrame
        )
    else
        self.wndShareFrame:SetSprite(ParaseleneWildRace.disabledToolbarButtonFrame)
    end
end

function ParaseleneWildRace.Client:OnBtnStartStop(wndHandler, wndControl, eMouseButton)
    if self.isStartStopEnabled then
        if self.isRaceInProgress then
            self:Stop()
            self:AdjustScrollbar()
        else
            self:Start()
        end
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    end
end

function ParaseleneWildRace.Client:OnBtnStartStopGenerateTooltip(wndHandler, wndControl, eToolTipType, x, y)
    self.btnStartStop:SetTooltip(
        (self.isRaceInProgress and "Stop the race") or
        "Start the race"
    )
end

function ParaseleneWildRace.Client:OnBtnStartStopMouseEnter(wndHandler, wndControl, x, y)
    if self.isStartStopEnabled then
        self.wndStartStopFrame:SetSprite(
            (self.isRaceInProgress and ParaseleneWildRace.pressedToolbarButtonFlybyFrame) or
            ParaseleneWildRace.enabledToolbarButtonFlybyFrame
        )
    else
        self.wndStartStopFrame:SetSprite(ParaseleneWildRace.disabledToolbarButtonFrame)
    end
end

function ParaseleneWildRace.Client:OnBtnStartStopMouseExit(wndHandler, wndControl, x, y)
    if self.isStartStopEnabled then
        self.wndStartStopFrame:SetSprite(
            (self.isRaceInProgress and ParaseleneWildRace.pressedToolbarButtonFrame) or
            ParaseleneWildRace.enabledToolbarButtonFrame
        )
    else
        self.wndStartStopFrame:SetSprite(ParaseleneWildRace.disabledToolbarButtonFrame)
    end
end

function ParaseleneWildRace.Client:OnBtnVisibleWaypoints(wndHandler, wndControl, eMouseButton)
    self.app.characterData.areWaypointsVisible = not self.app.characterData.areWaypointsVisible
    self:AdjustToolbar()
    Sound.Play(ParaseleneWildRace.buttonClickSound)
end

function ParaseleneWildRace.Client:OnBtnVisibleWaypointsGenerateTooltip(wndHandler, wndControl, eToolTipType, x, y)
    self.btnVisibleWaypoints:SetTooltip(
        (self.app.characterData.areWaypointsVisible and "Turn off waypoint markers and lines in the game world") or
        "Turn on waypoint markers and lines in the game world"
    )
end

function ParaseleneWildRace.Client:OnBtnVisibleWaypointsMouseEnter(wndHandler, wndControl, x, y)
    self.wndVisibleWaypointsFrame:SetSprite(
        (self.app.characterData.areWaypointsVisible and ParaseleneWildRace.pressedToolbarButtonFlybyFrame) or
        ParaseleneWildRace.enabledToolbarButtonFlybyFrame
    )
end

function ParaseleneWildRace.Client:OnBtnVisibleWaypointsMouseExit(wndHandler, wndControl, x, y)
    self.wndVisibleWaypointsFrame:SetSprite(
        (self.app.characterData.areWaypointsVisible and ParaseleneWildRace.pressedToolbarButtonFrame) or
        ParaseleneWildRace.enabledToolbarButtonFrame
    )
end

function ParaseleneWildRace.Client:OnBtnVisitHome(wndHandler, wndControl, eMouseButton)
    if self.isVisitHomeEnabled and self.visitZone then
        if self.visitZone == ParaseleneWildRaceInstance.name then
            HousingLib.RequestTakeMeHome()
        else
            HousingLib.RequestVisitPlayer(self.visitZone)
        end
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    end
end

function ParaseleneWildRace.Client:OnBtnVisitHomeMouseEnter(wndHandler, wndControl, x, y)
    self.wndVisitHomeFrame:SetSprite(
        (self.isVisitHomeEnabled and ParaseleneWildRace.enabledToolbarButtonFlybyFrame) or
        ParaseleneWildRace.disabledToolbarButtonFrame
    )
end

function ParaseleneWildRace.Client:OnBtnVisitHomeMouseExit(wndHandler, wndControl, x, y)
    self.wndVisitHomeFrame:SetSprite(
        (self.isVisitHomeEnabled and ParaseleneWildRace.enabledToolbarButtonFrame) or
        ParaseleneWildRace.disabledToolbarButtonFrame
    )
end

function ParaseleneWildRace.Client:OnEdtCourseNameEscape(wndHandler, wndControl)
    self.app.editor:OnEdtCourseNameEscape(wndHandler, wndControl)
end

function ParaseleneWildRace.Client:OnEdtCourseNameReturn(wndHandler, wndControl, strText)
    self.app.editor:OnEdtCourseNameReturn(wndHandler, wndControl, strText)
end

function ParaseleneWildRace.Client:OnEdtWaypointNameEscape(wndHandler, wndControl)
    self.app.editor:OnEdtWaypointNameEscape(wndHandler, wndControl)
end

function ParaseleneWildRace.Client:OnEdtWaypointNameReturn(wndHandler, wndControl, strText)
    self.app.editor:OnEdtWaypointNameReturn(wndHandler, wndControl, strText)
end

function ParaseleneWildRace.Client:OnRacePositionTimer()
    self.currentPosition = ParaseleneWildRace.Waypoint.GetCurrentPosition()
    if self.currentPosition  and self.isRaceInProgress and not self.isRaceFinished then
        if self.course.strictMode then
            -- In strict mode, the only hit test made is against the
            -- first un-visited, visitable waypoint in the course.
            local waypoint = self.course.waypoints[self:GetFirstVisitableIndex{skipVisited = true}]
            self:CheckPosition(self.currentPosition, waypoint)
        else
            -- In non-strict (loose) mode, hit tests are conducted against all waypoints in the course.
            for idx, waypoint in ipairs(self.course.waypoints) do
                self:CheckPosition(self.currentPosition, waypoint)
            end
        end
    end
end

function ParaseleneWildRace.Client:OnUpdateTimer()
    -- If OnRacePositionTimer() is not updating the current position, then update it here.
    if not self.racePositionTimer then
        self.currentPosition = ParaseleneWildRace.Waypoint.GetCurrentPosition()
    end

    -- Is the currently selected waypoint a home that we can visit?
    -- If so, remember where that is. Used by OnBtnVisitHome().
    if HousingLib.IsHousingWorld() and self.index then
        local waypoint = self.course.waypoints[self.index]
        self.isVisitHomeEnabled = waypoint.isVisitable and waypoint.inHousing
        self.visitZone = (self.isVisitHomeEnabled and waypoint.zone) or nil
    else
        self.isVisitHomeEnabled = false
        self.visitZone = nil
    end

    -- Adjust the visit home button.
    if self.isVisitHomeEnabled then
        self.wndVisitHomeFrame:SetSprite(
            (self.btnVisitHome:ContainsMouse() and ParaseleneWildRace.enabledToolbarButtonFlybyFrame) or
            ParaseleneWildRace.enabledToolbarButtonFrame
        )
        self.wndVisitHomeDisabled:SetSprite("")
    else
        self.wndVisitHomeFrame:SetSprite(ParaseleneWildRace.disabledToolbarButtonFrame)
        self.wndVisitHomeDisabled:SetSprite(ParaseleneWildRace.noIcon)
    end
    self.btnVisitHome:Enable(self.isVisitHomeEnabled)

    if self.isRaceInProgress and not self.isRaceFinished then
        self:UpdateRaceClock()
    end

    self.app.about:SetContent(self.currentPosition)

    local visibleWaypoint = self.app.characterData.areWaypointsVisible and
        self.index and self.course.waypoints[self.index]
    self.app.trackMaster:Connect()
    self.app.trackMaster:SetTarget(visibleWaypoint, self.currentPosition)
    self.app.worldMarker:SetTarget(visibleWaypoint, self.currentPosition)
end

function ParaseleneWildRace.Client:OnWaypointSelected(
        wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY, bDoubleClick, bStopPropagation)
    local index = wndHandler:GetData()
    if index and index ~= self.index then
        self:SetIndex(index)
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    end
end

function ParaseleneWildRace.Client:OnWindowMove(wndHandler, wndControl, nOldLeft, nOldTop, nOldRight, nOldBottom)
    self:OnWindowMoveOrSizeChanged()
end

function ParaseleneWildRace.Client:OnWindowMoveOrSizeChanged()
    -- Keep self.app.characterData.clientWindowInfo up to date with the window's position and size.
    local left, top, right, bottom = self.wndClient:GetRect()
    self.app.characterData.clientWindowInfo.left = left or 0
    self.app.characterData.clientWindowInfo.top = top or 0
    self.app.characterData.clientWindowInfo.width = (right and (right - left)) or
        ParaseleneWildRace.minClientWindowWidth
    self.app.characterData.clientWindowInfo.height = (bottom and (bottom - top)) or
        ParaseleneWildRace.minClientWindowHeight
end

function ParaseleneWildRace.Client:OnWindowSizeChanged(wndHandler, wndControl)
    self:OnWindowMoveOrSizeChanged()
end

function ParaseleneWildRace.Client:SetCourse(course)
    self:ClearCourse()
    if course then
        self.course = ParaseleneWildRace.Course:new(course) -- Deep copy.
        self.edtCourseName:SetText(self.course.name)
        if #self.course.waypoints > 0 then
            for idx, waypoint in ipairs(self.course.waypoints) do
                waypoint.wndWaypointListItem = Apollo.LoadForm(
                    self.app.doc, "wndWaypointListItem", self.wndWaypointList, self
                )
                if not waypoint.wndWaypointListItem then
                    local message = ParaseleneWildRace.appName ..
                        ": Error: Could not load a waypoint list item window."
                    ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_System, message)
                    Apollo.AddAddonErrorText(ParaseleneWildRaceInstance, message)

                    self:ClearCourse()
                    return
                end

                -- Index of the associated waypoint in course.
                waypoint.wndWaypointListItem:SetData(idx)

                -- Cube for position check.
                waypoint.isVisitable = waypoint:IsVisitable()
                if waypoint.isVisitable then
                    waypoint.cube = ParaseleneWildRace.Cube.GetPosition(waypoint.point)
                end

                -- Waypoint status icon (visitable or not).
                waypoint.wndStatus = waypoint.wndWaypointListItem:FindChild("wndStatus")
                if waypoint.isVisitable then
                    waypoint.wndStatus:SetSprite("")
                    waypoint.wndStatus:SetTooltip("")
                else
                    waypoint.wndStatus:SetSprite(ParaseleneWildRace.noIcon)
                    waypoint.wndStatus:SetTooltip("The waypoint cannot be visited")
                end

                -- Waypoint house icon.
                local wndIcon = waypoint.wndWaypointListItem:FindChild("wndIcon")
                if waypoint.inHousing then
                    wndIcon:SetSprite(ParaseleneWildRace.houseIcon)
                    wndIcon:SetTooltip("The waypoint is in a home")
                else
                    wndIcon:SetSprite("")
                    wndIcon:SetTooltip("")
                end

                -- Waypoint name and zone.
                local edtName = waypoint.wndWaypointListItem:FindChild("edtName")
                edtName:SetText(waypoint.name)
                edtName:SetTooltip(waypoint.name)
                local edtZone = waypoint.wndWaypointListItem:FindChild("edtZone")
                edtZone:SetText(waypoint.zone)
                edtZone:SetTooltip(waypoint.zone)
            end
            self.wndWaypointList:ArrangeChildrenVert()
        end

        self:SetIndex(((#self.course.waypoints > 0) and 1) or nil)

        self:AdjustToolbar()
        self:AdjustStatusbar()
        if self.app.editor then
            self.app.editor:AdjustRaceTools()
        end
        if self.app.files then
            self.app.files:AdjustTools()
        end
    end
end

function ParaseleneWildRace.Client:SetIndex(index)
    -- Remove the border around the currently selected waypoint in the list.
    if self.index then
        self.course.waypoints[self.index].wndWaypointListItem:SetSprite("")
    end

    if self.course and (#self.course.waypoints > 0) then
        self.index = (index and (index >= 1) and (index <= #self.course.waypoints) and index) or 1

        -- Set the border around the newly selected waypoint in the list.
        self.course.waypoints[self.index].wndWaypointListItem:SetSprite(ParaseleneWildRace.selectedItemBorder)

        self:SetWaypointDetails(self.course.waypoints[self.index])
    else
        self.index = nil
        self:SetWaypointDetails(nil)
    end

    if self.app.editor then
        self.app.editor:AdjustWaypointTools()
    end
end

function ParaseleneWildRace.Client:SetScrollPosition(position)
    local windowHeight = self.wndWaypointList:GetHeight()
    local listHeight = #self.course.waypoints * ParaseleneWildRace.waypointListItemHeight
    if (position + windowHeight) > listHeight then
        position = listHeight - windowHeight
    end
    if position < 0 then
        position = 0
    end
    if position ~= self.wndWaypointList:GetVScrollPos() then
        self.wndWaypointList:SetVScrollPos(position)
    end
end

function ParaseleneWildRace.Client:SetWaypointDetails(waypoint)
    local headers, details = "", ""

    if waypoint then
        if waypoint.inHousing then
            headers = "Home:\nPosition:"
        else
            headers = "Zone:\nPosition:"
        end
        details = waypoint.zone .. ", " .. waypoint.subZone .. "\n" ..
            string.format("%.0f, %.0f, %.0f", waypoint.point.x, waypoint.point.z, waypoint.point.y)
    end

    self.edtWaypointName:SetText((waypoint and waypoint.name) or "")
    self.wndWaypointDetailsHeaders:SetText(headers)
    self.edtWaypointDetails:SetText(details)

    -- Keep the waypoint name from unwanted highlighted.
    -- Unwanted highlighting occurs when text is set in an editbox that was once editable (non-ReadOnly).
    self.edtWaypointName:ClearFocus()
end

function ParaseleneWildRace.Client:SetWindowLock(isLocked)
    if isLocked then
        self.app.characterData.clientWindowInfo.locked = true
        self.wndClient:SetStyle("Moveable", false)
        self.wndClient:SetStyle("Sizable", false)
        self.wndLockIcon:SetSprite(ParaseleneWildRace.lockEnabledIcon)
    else
        self.app.characterData.clientWindowInfo.locked = false
        self.wndClient:SetStyle("Moveable", true)
        self.wndClient:SetStyle("Sizable", true)
        self.wndLockIcon:SetSprite(ParaseleneWildRace.lockDisabledIcon)
    end
end

function ParaseleneWildRace.Client:ShowWindow()
    if not self.wndClient:IsVisible() then
        -- When the client window closes, the satallite windows also close.
        -- When the client window re-opens, also re-open the satallite windows.
        if self.wasEditorVisibleLastTimeClientWasVisible then
            self.app.editor:ShowWindow()
        end
        if self.wasFilesVisibleLastTimeClientWasVisible then
            self.app.files:ShowWindow()
        end
        if self.wasShareVisibleLastTimeClientWasVisible then
            self.app.share:ShowWindow()
        end

        self.wndClient:Invoke()
    end
end

function ParaseleneWildRace.Client:Start()
    -- The course must have at least one visitable waypoint in it.
    local firstVisitableIndex = self:GetFirstVisitableIndex()
    if firstVisitableIndex then
        self:Stop() -- If a race is in progress, then this is a restart.

        local message = "Race started"
        ChatSystemLib.PostOnChannel(
            ChatSystemLib.ChatChannel_System,
            ParaseleneWildRace.appName .. ": " .. message .. "."
        )
        self.Broadcast(message, ParaseleneWildRace.broadcastBlue)

        self.raceStartClock = os.clock()
        self.isRaceInProgress = true
        self.isRaceFinished = false

        self.app.about:HideWindow()
        self.app.editor:HideWindow()
        self.app.files:HideWindow()
        self.app.options:HideWindow()
        self.app.share:HideWindow()

        self:SetIndex(firstVisitableIndex)
        self:AdjustScrollbar()

        self:AdjustToolbar()

        -- Only needed if the status bar is updated to a custom message while a race is not in progress.
        -- self:AdjustStatusbar()

        if not self.racePositionTimer then
            self.racePositionTimer = ApolloTimer.Create(
                RACE_POSITION_TIMER_TICK_INTERVAL_SECS, true, "OnRacePositionTimer", self
            )
        end
    else
        local message = "Race cannot start; no visitable waypoints"
        ChatSystemLib.PostOnChannel(
            ChatSystemLib.ChatChannel_System,
            ParaseleneWildRace.appName .. ": " .. message .. "."
        )
        self.Broadcast(message, ParaseleneWildRace.broadcastRed)
    end
end

function ParaseleneWildRace.Client:Stop()
    if self.isRaceInProgress then
        local message = "Race stopped"
        ChatSystemLib.PostOnChannel(
            ChatSystemLib.ChatChannel_System,
            ParaseleneWildRace.appName .. ": " .. message .. "."
        )
        self.Broadcast(message, ParaseleneWildRace.broadcastRed)
    end

    self.isRaceInProgress = false
    self.isRaceFinished = false

    if self.racePositionTimer then
        self.racePositionTimer:Stop()
        self.racePositionTimer = nil
    end

    if self.course then
        for idx, waypoint in ipairs(self.course.waypoints) do
            if waypoint.isVisitable and waypoint.hasBeenVisited then
                waypoint.hasBeenVisited = false
                waypoint.wndStatus:SetSprite("") -- Clear the checkmark on the waypoint in the list.
                waypoint.wndStatus:SetTooltip("")
            end
        end
    end

    self:SetIndex((self.course and (#self.course.waypoints > 0) and 1) or nil)

    self:AdjustToolbar()
    self:AdjustStatusbar() -- In case the race "Finished" or "Expired" message is present.
end

function ParaseleneWildRace.Client:ToggleWindowVisibility()
    if self.wndClient:IsVisible() then
        self:HideWindow()
    else
        self:ShowWindow()
    end
end

function ParaseleneWildRace.Client:UpdateRaceClock()
    local currentClock = os.clock()
    local elapsed = (((currentClock - self.raceStartClock) >= 0) and (currentClock - self.raceStartClock)) or 0
    if self.course.timedMode then
        -- Toolbar timer; remaining time.
        local remaining = (((self.course.timeout - elapsed) >= 0) and (self.course.timeout - elapsed)) or 0
        self.wndTimer:SetText(self.FormatTime(remaining))

        if remaining == 0 then
            -- No more time left. We have finished the race and lost.
            self.isRaceFinished = true

            self.wndStatus:SetText("Expired:")

            local message = "Race timer expired"
            ChatSystemLib.PostOnChannel(
                ChatSystemLib.ChatChannel_System,
                ParaseleneWildRace.appName .. ": " .. message .. "."
            )
            self.Broadcast(message, ParaseleneWildRace.broadcastRed)
        end
    else
        -- Toolbar timer; elapsed time.
        self.wndTimer:SetText(self.FormatTime(elapsed))
    end
end
