-- Paraselene's WildRace
-- Version 0.5.2

ParaseleneWildRace.Course = {}
 
function ParaseleneWildRace.Course:new(o)
    if type(o) ~= "table" then
        return nil
    end

    local i = {}
    setmetatable(i, self)
    self.__index = self 

    i.name = ((type(o.name) == "string") and o.name) or nil
    i.uid = ((type(o.uid) == "number") and o.uid) or 0 -- Zero means "unassigned" to a folder.

    -- Does not vary, so, no need for check or persistence.
    i.folderItemType = ParaseleneWildRace.folderItemTypes.course

    -- i.waypoints cannot be nil, even if the o.waypoints is nil.
    i.waypoints = {}
    if o.waypoints and (type(o.waypoints) == "table") then
        for idx, waypoint in ipairs(o.waypoints) do
            i.waypoints[idx] = ParaseleneWildRace.Waypoint:new(waypoint)
            if not i.waypoints[idx] then
                return nil
            end
        end
    end

    if type(o.strictMode) == "boolean" then
        i.strictMode = o.strictMode
    end

    if type(o.timedMode) == "boolean" then
        i.timedMode = o.timedMode
    end

    i.timeout = ((type(o.timeout) == "number") and o.timeout) or nil

    return (i:IsValid() and i) or nil
end

function ParaseleneWildRace.Course:Append(waypoint)
    table.insert(self.waypoints, waypoint)
end

function ParaseleneWildRace.Course:Delete(index)
    table.remove(self.waypoints, index)
end

function ParaseleneWildRace.Course:Insert(index, waypoint)
    table.insert(self.waypoints, index, waypoint)
end

function ParaseleneWildRace.Course:IsValid()
    -- self.uid does not need to be checked. It will always exist.
    -- self.waypoints does not need to be checked. It will always exist.
    return (self.name ~= nil) and
           (self.strictMode ~= nil) and
           (self.timedMode ~= nil) and
           (self.timeout ~= nil)
end

function ParaseleneWildRace.Course:Save()
    local waypoints = {}
    for idx, waypoint in ipairs(self.waypoints) do
        waypoints[idx] = waypoint:Save()
    end
    return {
        name = self.name,
        uid = self.uid,
        waypoints = waypoints,
        strictMode = self.strictMode,
        timedMode = self.timedMode,
        timeout = self.timeout
    }
end

function ParaseleneWildRace.Course:ShiftEarlier(index)
    self.waypoints[index - 1], self.waypoints[index] = self.waypoints[index], self.waypoints[index - 1]
end

function ParaseleneWildRace.Course:ShiftLater(index)
    self.waypoints[index], self.waypoints[index + 1] = self.waypoints[index + 1], self.waypoints[index]
end
