-- Paraselene's WildRace
-- Version 0.5.2

local CUBE_WIDTH = 10
local CUBE_RADIUS = CUBE_WIDTH / 2

ParaseleneWildRace.Cube = {}
 
function ParaseleneWildRace.Cube:new(o)
    if type(o) ~= "table" then
        return nil
    end

    local i = {}
    setmetatable(i, self)
    self.__index = self 

    i.pointLow = ParaseleneWildRace.Point:new(o.pointLow)
    i.pointHigh = ParaseleneWildRace.Point:new(o.pointHigh) -- Exclusive; outside of the cube.

    return (i:IsValid() and i) or nil
end

function ParaseleneWildRace.Cube.GetPosition(point)
    local pointLow = ParaseleneWildRace.Point:new{
        x = point.x - CUBE_RADIUS,
        y = point.y - CUBE_RADIUS,
        z = point.z - CUBE_RADIUS
    }
    local pointHigh = ParaseleneWildRace.Point:new{
        x = pointLow.x + CUBE_WIDTH,
        y = pointLow.y + CUBE_WIDTH,
        z = pointLow.z + CUBE_WIDTH
    }
    return ParaseleneWildRace.Cube:new{
        pointLow = pointLow,
        pointHigh = pointHigh
    }
end

function ParaseleneWildRace.Cube:InRange(point)
    return (point.x >= self.pointLow.x) and (point.x < self.pointHigh.x) and
           (point.y >= self.pointLow.y) and (point.y < self.pointHigh.y) and
           (point.z >= self.pointLow.z) and (point.z < self.pointHigh.z)
end

function ParaseleneWildRace.Cube:IsValid()
    return (self.pointLow ~= nil) and
           (self.pointHigh ~= nil)
end

function ParaseleneWildRace.Cube:Save()
    return {
        pointLow = self.pointLow:Save(),
        pointHigh = self.pointHigh:Save()
    }
end
