-- Paraselene's WildRace
-- Version 0.5.2

require "Apollo"
require "Window"
 
ParaseleneWildRace.Editor = {}

function ParaseleneWildRace.Editor:new(app)
    local i = {}
    setmetatable(i, self)
    self.__index = self

    i.app = app

    i.wndEditor = Apollo.LoadForm(i.app.doc, "wndEditor", nil, i)
    if not i.wndEditor then
        local message = ParaseleneWildRace.appName ..
            ": Error: Could not load the \"Editor\" window."
        ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_System, message)
        Apollo.AddAddonErrorText(ParaseleneWildRaceInstance, message)
        return nil
    end

    local wndTitle = i.wndEditor:FindChild("wndTitle")
    i.btnAbout = wndTitle:FindChild("btnAbout")
    i.btnLock = wndTitle:FindChild("btnLock")
    i.wndLockIcon = i.btnLock:FindChild("wndIcon")

    local wndTools = i.wndEditor:FindChild("wndTools")

    local wndCourseTools = wndTools:FindChild("wndCourseTools")
    i.btnCreate = wndCourseTools:FindChild("btnCreate")
    i.btnCreate:SetData(ParaseleneWildRace.Button:new{
        button = i.btnCreate,
        style = ParaseleneWildRace.buttonStyles.red,
        enabled = true
    })

    local wndWaypointTools = wndTools:FindChild("wndWaypointTools")
    i.btnAdd = wndWaypointTools:FindChild("btnAdd")
    i.btnAdd:SetData(ParaseleneWildRace.Button:new{
        button = i.btnAdd,
        enabled = true
    })
    i.btnMoveUp = wndWaypointTools:FindChild("btnMoveUp")
    i.btnMoveUp:SetData(ParaseleneWildRace.Button:new{
        button = i.btnMoveUp
    })
    i.btnMoveDown = wndWaypointTools:FindChild("btnMoveDown")
    i.btnMoveDown:SetData(ParaseleneWildRace.Button:new{
        button = i.btnMoveDown
    })
    i.btnCopy = wndWaypointTools:FindChild("btnCopy")
    i.btnCopy:SetData(ParaseleneWildRace.Button:new{
        button = i.btnCopy
    })
    i.btnPaste = wndWaypointTools:FindChild("btnPaste")
    i.btnPaste:SetData(ParaseleneWildRace.Button:new{
        button = i.btnPaste
    })
    i.btnDelete = wndWaypointTools:FindChild("btnDelete")
    i.btnDelete:SetData(ParaseleneWildRace.Button:new{
        button = i.btnDelete,
        style = ParaseleneWildRace.buttonStyles.red
    })

    local wndRaceTools = wndTools:FindChild("wndRaceTools")
    i.wndStrict = wndRaceTools:FindChild("wndStrict")
    i.btnStrict = i.wndStrict:FindChild("btnStrict")
    i.wndStrictIcon = i.btnStrict:FindChild("wndIcon")
    i.btnTimed = wndRaceTools:FindChild("btnTimed")
    i.wndTimedIcon = i.btnTimed:FindChild("wndIcon")
    local wndTimeout = wndRaceTools:FindChild("wndTimeout")
    i.wndTimeoutLabel = wndTimeout:FindChild("wndLabel")
    i.edtTimeout = wndTimeout:FindChild("wndTimeout"):FindChild("edtTimeout")

    wndTitle:SetText(ParaseleneWildRace.appName .. " Editor")

    local width = ParaseleneWildRace.editorWindowWidth
    local height = ParaseleneWildRace.editorWindowHeight
    i.wndEditor:Move(
        i.app.characterData.editorWindowInfo.left,
        i.app.characterData.editorWindowInfo.top,
        width,
        height
    )
    i:SetWindowLock(i.app.characterData.editorWindowInfo.locked)

    i.wndEditor:SetSprite(ParaseleneWildRace.backgroundSprites[i.app.characterData.uiOptions.background])
    i.wndEditor:Show(false, true)

    return i
end

function ParaseleneWildRace.Editor:AddWaypoint()
    local course, index = self:GetCourse()
    local waypoint = ParaseleneWildRace.Waypoint.GetCurrentPosition()
    if waypoint then
        if index then
            course:Insert(index, waypoint)
        else
            course:Append(waypoint)
        end
    end
    self:SetCourseAndIndex(course, index or 1) -- Move the selection to the newly added waypoint.
end

--[[
function ParaseleneWildRace.Editor:AdjustCourseTools()
    -- The create new course button is already enabled.
end
--]]

function ParaseleneWildRace.Editor:AdjustRaceTools()
    -- Check or uncheck the strict mode checkbox.
    if self.app.client and self.app.client.course and self.app.client.course.strictMode then
        self.btnStrict:SetData(true) -- Checked. Used by mouse enter/exit handlers.
        self.wndStrictIcon:SetSprite(
            (self.btnStrict:ContainsMouse() and ParaseleneWildRace.checkedCheckboxFlybyIcon) or
            ParaseleneWildRace.checkedCheckboxIcon
        )
    else
        self.btnStrict:SetData(false) -- Unchecked. Used by mouse enter/exit handlers.
        self.wndStrictIcon:SetSprite(
            (self.btnStrict:ContainsMouse() and ParaseleneWildRace.uncheckedCheckboxFlybyIcon) or
            ParaseleneWildRace.uncheckedCheckboxIcon
        )
    end

    self.isTimeoutEnabled = self.app.client and self.app.client.course and self.app.client.course.timedMode

    -- Check or uncheck the timed mode checkbox.
    if self.isTimeoutEnabled then
        self.btnTimed:SetData(true) -- Checked. Used by mouse enter/exit handlers.
        self.wndTimedIcon:SetSprite(
            (self.btnTimed:ContainsMouse() and ParaseleneWildRace.checkedCheckboxFlybyIcon) or
            ParaseleneWildRace.checkedCheckboxIcon
        )
    else
        self.btnTimed:SetData(false) -- Unchecked. Used by mouse enter/exit handlers.
        self.wndTimedIcon:SetSprite(
            (self.btnTimed:ContainsMouse() and ParaseleneWildRace.uncheckedCheckboxFlybyIcon) or
            ParaseleneWildRace.uncheckedCheckboxIcon
        )
    end

    -- Color the timeout label and editbox as either enabled or disabled.
    if self.isTimeoutEnabled then
        self.wndTimeoutLabel:SetTextColor(ApolloColor.new(ParaseleneWildRace.enabledTextColor))
        self.edtTimeout:SetTextColor(ApolloColor.new(ParaseleneWildRace.enabledTextColor))
    else
        self.wndTimeoutLabel:SetTextColor(ApolloColor.new(ParaseleneWildRace.disabledTextColor))
        self.edtTimeout:SetTextColor(ApolloColor.new(ParaseleneWildRace.disabledTextColor))
    end

    -- Enable or disable the timeout editbox.
    if self.isTimeoutEnabled then
        self.edtTimeout:SetStyle("IgnoreMouse", false)
        self.edtTimeout:SetStyleEx("ReadOnly", false)
    else
        self.edtTimeout:SetStyle("IgnoreMouse", true)
        self.edtTimeout:SetStyleEx("ReadOnly", true)
        self.edtTimeout:ClearFocus() -- Keep the timeout from unwanted highlighted.
    end

    -- Set the timeout. Use the client's FormatTime().
    if self.app.client and self.app.client.course then
        self.edtTimeout:SetText(self.app.client.FormatTime(self.app.client.course.timeout))
    else
        self.edtTimeout:SetText(self.app.client.FormatTime(0))
    end
end

function ParaseleneWildRace.Editor:AdjustWaypointTools()
    -- The add button is already enabled.

    -- Enable or disable the move up button.
    if self.app.client and self.app.client.index and (self.app.client.index > 1) then
        self.btnMoveUp:GetData():Enable()
    else
        self.btnMoveUp:GetData():Disable()
    end

    -- Enable or disable the move down button.
    if self.app.client and self.app.client.index and (self.app.client.index < #self.app.client.course.waypoints) then
        self.btnMoveDown:GetData():Enable()
    else
        self.btnMoveDown:GetData():Disable()
    end

    -- Enable or disable the copy button.
    if self.app.client and self.app.client.index then
        self.btnCopy:GetData():Enable()
    else
        self.btnCopy:GetData():Disable()
    end

    -- Enable or disable the paste button.
    if self.clipboard then
        self.btnPaste:GetData():Enable()
    else
        self.btnPaste:GetData():Disable()
    end

    -- Enable or disable the delete button.
    if self.app.client and self.app.client.index then
        self.btnDelete:GetData():Enable()
    else
        self.btnDelete:GetData():Disable()
    end
end

function ParaseleneWildRace.Editor:DisableEditboxes()
    -- Disable editing the course name.
    self.app.client.edtCourseName:SetStyle("IgnoreMouse", true)
    self.app.client.edtCourseName:SetStyleEx("ReadOnly", true)

    -- Reset the course name to clear out any unfinished editing.
    local name = (self.app.client.course and self.app.client.course.name) or ""
    self.app.client.edtCourseName:SetText(name)
    self.app.client.edtCourseName:ClearFocus() -- Keep the timeout from unwanted highlighted.

    -- Disable editing the waypoint name.
    self.app.client.edtWaypointName:SetStyle("IgnoreMouse", true)
    self.app.client.edtWaypointName:SetStyleEx("ReadOnly", true)

    -- Reset the waypoint name to clear out any unfinished editing.
    name = (self.app.client.index and self.app.client.course.waypoints[self.app.client.index].name) or ""
    self.app.client.edtWaypointName:SetText(name)
    self.app.client.edtWaypointName:ClearFocus() -- Keep the timeout from unwanted highlighted.

    -- Reset the timeout to clear out any unfinished editing.
    local timeoutText = (self.app.client.course and self.app.client.FormatTime(self.app.client.course.timeout)) or ""
    self.edtTimeout:SetText(timeoutText)
    self.edtTimeout:ClearFocus() -- Keep the timeout from unwanted highlighted.
end

function ParaseleneWildRace.Editor:EnableEditboxes()
    -- Enable editing the course name.
    self.app.client.edtCourseName:SetStyle("IgnoreMouse", false)
    self.app.client.edtCourseName:SetStyleEx("ReadOnly", false)

    -- Enable editing the waypoint name.
    self.app.client.edtWaypointName:SetStyle("IgnoreMouse", false)
    self.app.client.edtWaypointName:SetStyleEx("ReadOnly", false)
end

function ParaseleneWildRace.Editor.FormatTimeout(timeoutString)
    -- If all else fails, treat as zero seconds.
    local timeoutNumber = 0

    -- Try to read the string as hours, minutes, and seconds.
    local hours, minutes, seconds = string.match(timeoutString, "^%s*(%d+):(%d+):(%d+)%s*$")
    if hours and minutes and seconds then
        timeoutNumber = (tonumber(hours) * 3600) + (tonumber(minutes) * 60) + tonumber(seconds)
    end

    -- Try to read the string as minutes and seconds.
    minutes, seconds = string.match(timeoutString, "^%s*(%d+):(%d+)%s*$")
    if minutes and seconds then
        timeoutNumber = (tonumber(minutes) * 60) + tonumber(seconds)
    end

    -- Try to read the string as a single number of seconds.
    seconds = string.match(timeoutString, "^%s*(%d+)%s*$")
    if seconds then
        timeoutNumber = tonumber(seconds)
    end

    -- Bounds check.
    return ((timeoutNumber > ParaseleneWildRace.maxRaceTime) and ParaseleneWildRace.maxRaceTime) or timeoutNumber
end

function ParaseleneWildRace.Editor:GetCourse()
    local course = ParaseleneWildRace.Course:new(self.app.client.course)
    local index = self.app.client.index

    -- If no course is present, make an empty one.
    if not course then
        course = ParaseleneWildRace.Course:new{
            name = "New Course",
            strictMode = false,
            timedMode = false,
            timeout = 0
        }
    end

    return course, index
end

function ParaseleneWildRace.Editor:HideWindow()
    if self.wndEditor:IsVisible() then
        self:DisableEditboxes()
        self.wndEditor:Close()
        self.app.client.wndEditorFrame:SetSprite(
            (self.app.client.btnEditor:ContainsMouse() and ParaseleneWildRace.enabledToolbarButtonFlybyFrame) or
            ParaseleneWildRace.enabledToolbarButtonFrame
        )
    end
end

function ParaseleneWildRace.Editor:OnBtnAbout(wndHandler, wndControl, eMouseButton)
    self.app.about:ToggleWindowVisibility()
    Sound.Play(ParaseleneWildRace.buttonClickSound)
end

function ParaseleneWildRace.Editor:OnBtnAboutGenerateTooltip(wndHandler, wndControl, eToolTipType, x, y)
    self.btnAbout:SetTooltip(
        (self.app.about.wndAbout:IsVisible() and "Close the \"About\" window") or
        "Open the \"About\" window"
    )
end

function ParaseleneWildRace.Editor:OnBtnAdd(wndHandler, wndControl, eMouseButton)
    self:AddWaypoint()
    Sound.Play(ParaseleneWildRace.buttonClickSound)
end

function ParaseleneWildRace.Editor:OnBtnClose(wndHandler, wndControl, eMouseButton)
    self:HideWindow()
    Sound.Play(ParaseleneWildRace.buttonClickSound)
end

function ParaseleneWildRace.Editor:OnBtnCopy(wndHandler, wndControl, eMouseButton)
    if self.btnCopy:GetData():IsEnabled() then
        -- Deep copy.
        self.clipboard = ParaseleneWildRace.Waypoint:new(self.app.client.course.waypoints[self.app.client.index])
        self:AdjustWaypointTools()
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    end
end

function ParaseleneWildRace.Editor:OnBtnCreate(wndHandler, wndControl, eMouseButton)
    if self.btnCreate:GetData():IsEnabled() then
        self.app.client:SetCourse(ParaseleneWildRace.Course:new{
            name = "New Course",
            strictMode = false,
            timedMode = false,
            timeout = 0
        })
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    end
end

function ParaseleneWildRace.Editor:OnBtnDelete(wndHandler, wndControl, eMouseButton)
    if self.btnDelete:GetData():IsEnabled() then
        local course, index = self:GetCourse()
        course:Delete(index)
        local scrollPosition = self.app.client:GetScrollPosition()
        self.app.client:SetCourse(course)

        -- If there are two or more waypoints still in the course, then adjust the selection.
        -- The selection is already valid if the course contains one or no waypoints.
        if #self.app.client.course.waypoints >= 2 then
            if index > #self.app.client.course.waypoints then -- If we deleted the tail waypoint in the course ...
                index = #self.app.client.course.waypoints -- ... then point to the new tail waypoint in the course.
            end
            self.app.client:SetIndex(index)
        end
        self.app.client:SetScrollPosition(scrollPosition)
        self.app.client:AdjustScrollbar()
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    end
end

function ParaseleneWildRace.Editor:OnBtnLock(wndHandler, wndControl, eMouseButton)
    self:SetWindowLock(not self.app.characterData.editorWindowInfo.locked)
    Sound.Play(ParaseleneWildRace.buttonClickSound)
end

function ParaseleneWildRace.Editor:OnBtnLockGenerateTooltip(wndHandler, wndControl, eToolTipType, x, y)
    self.btnLock:SetTooltip(
        (self.app.characterData.editorWindowInfo.locked and "Unlock the window so that it can be moved") or
        "Lock the window so that it cannot be moved"
    )
end

function ParaseleneWildRace.Editor:OnBtnMoveDown(wndHandler, wndControl, eMouseButton)
    if self.btnMoveDown:GetData():IsEnabled() then
        local course, index = self:GetCourse()
        course:ShiftLater(index)
        self:SetCourseAndIndex(course, index + 1) -- Keep the selection on the moving waypoint.
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    end
end

function ParaseleneWildRace.Editor:OnBtnMoveUp(wndHandler, wndControl, eMouseButton)
    if self.btnMoveUp:GetData():IsEnabled() then
        local course, index = self:GetCourse()
        course:ShiftEarlier(index)
        self:SetCourseAndIndex(course, index - 1) -- Keep the selection on the moving waypoint.
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    end
end

function ParaseleneWildRace.Editor:OnBtnPaste(wndHandler, wndControl, eMouseButton)
    if self.btnPaste:GetData():IsEnabled() then
        local course, index = self:GetCourse()
        if index then
            -- Insert a deep copy.
            course:Insert(index, ParaseleneWildRace.Waypoint:new(self.clipboard))
        else
            -- Append a deep copy.
            course:Append(ParaseleneWildRace.Waypoint:new(self.clipboard))
        end
        self:SetCourseAndIndex(course, index or 1) -- Keep the selection on the pasted waypoint.
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    end
end

function ParaseleneWildRace.Editor:OnBtnStrict(wndHandler, wndControl, eMouseButton)
    local course, index = self:GetCourse()
    course.strictMode = not course.strictMode
    self:SetCourseAndIndex(course, index) -- Preserve the selection.
    Sound.Play(ParaseleneWildRace.buttonClickSound)
end

function ParaseleneWildRace.Editor:OnBtnTimed(wndHandler, wndControl, eMouseButton)
    local course, index = self:GetCourse()
    course.timedMode = not course.timedMode
    self:SetCourseAndIndex(course, index) -- Preserve the selection.
    Sound.Play(ParaseleneWildRace.buttonClickSound)
end

function ParaseleneWildRace.Editor:OnBtnTimedGenerateTooltip(wndHandler, wndControl, eToolTipType, x, y)
    self.btnTimed:SetTooltip(
        (self.app.client.course.timedMode and
         "Set the course so that, when racing, it can be completed is any amount of time") or
         "Set the course so that, when racing, it must be completed in the specified amount of time"
    )
end

function ParaseleneWildRace.Editor:OnButtonMouseEnter(wndHandler, wndControl, x, y)
    wndHandler:GetData():OnMouseEnter()
end

function ParaseleneWildRace.Editor:OnButtonMouseExit(wndHandler, wndControl, x, y)
    wndHandler:GetData():OnMouseExit()
end

function ParaseleneWildRace.Editor:OnCheckboxMouseEnter(wndHandler, wndControl, x, y)
    local wndIcon = wndHandler:FindChild("wndIcon")
    local checked = wndHandler:GetData()
    if checked then
        wndIcon:SetSprite(ParaseleneWildRace.checkedCheckboxFlybyIcon)
    else
        wndIcon:SetSprite(ParaseleneWildRace.uncheckedCheckboxFlybyIcon)
    end
end

function ParaseleneWildRace.Editor:OnCheckboxMouseExit(wndHandler, wndControl, x, y)
    local wndIcon = wndHandler:FindChild("wndIcon")
    local checked = wndHandler:GetData()
    if checked then
        wndIcon:SetSprite(ParaseleneWildRace.checkedCheckboxIcon)
    else
        wndIcon:SetSprite(ParaseleneWildRace.uncheckedCheckboxIcon)
    end
end

function ParaseleneWildRace.Editor:OnEdtCourseNameEscape(wndHandler, wndControl)
    -- Reset the course name to clear out any unfinished editing.
    local name = (self.app.client.course and self.app.client.course.name) or ""
    self.app.client.edtCourseName:SetText(name)
end

function ParaseleneWildRace.Editor:OnEdtCourseNameReturn(wndHandler, wndControl, strText)
    local newText = string.gsub(strText, "^%s*(.-)%s*$", "%1") -- Strip leading and trailing spaces.
    if newText ~= "" then
        local course, index = self:GetCourse()
        course.name = newText
        self:SetCourseAndIndex(course, index) -- Preserve the selection.
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    else
        self:OnEdtCourseNameEscape(wndHandler, wndControl)
    end
end

function ParaseleneWildRace.Editor:OnEdtTimeoutEscape(wndHandler, wndControl)
    -- Reset the timeout to clear out any unfinished editing.
    self.edtTimeout:SetText(self.app.client.FormatTime(self.app.client.course.timeout))
end

function ParaseleneWildRace.Editor:OnEdtTimeoutReturn(wndHandler, wndControl, strText)
    self.app.client.course.timeout = self.FormatTimeout(strText)
    self.edtTimeout:SetText(self.app.client.FormatTime(self.app.client.course.timeout))
    self.app.client:AdjustStatusbar()
    Sound.Play(ParaseleneWildRace.buttonClickSound)
end

function ParaseleneWildRace.Editor:OnEdtWaypointNameEscape(wndHandler, wndControl)
    -- Reset the waypoint name to clear out any unfinished editing.
    local name = (self.app.client.index and self.app.client.course.waypoints[self.app.client.index].name) or ""
    self.app.client.edtWaypointName:SetText(name)
end

function ParaseleneWildRace.Editor:OnEdtWaypointNameReturn(wndHandler, wndControl, strText)
    local newText = string.gsub(strText, "^%s*(.-)%s*$", "%1") -- Strip leading and trailing spaces.
    if newText ~= "" then
        -- If no waypoints exist, create a new one so that we can name it.
        if not self.app.client.index then
            self:AddWaypoint()
        end

        local course, index = self:GetCourse()
        course.waypoints[index].name = newText
        self:SetCourseAndIndex(course, index) -- Preserve the selection.
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    else
        self:OnEdtWaypointNameEscape(wndHandler, wndControl)
    end
end

function ParaseleneWildRace.Editor:OnWindowMove(wndHandler, wndControl, nOldLeft, nOldTop, nOldRight, nOldBottom)
    -- Keep self.app.characterData.editorWindowInfo up to date with the window's position.
    local left, top, right, bottom = self.wndEditor:GetRect()
    self.app.characterData.editorWindowInfo.left = left or 0
    self.app.characterData.editorWindowInfo.top = top or 0
end

function ParaseleneWildRace.Editor:OnWndStrictGenerateTooltip(wndHandler, wndControl, eToolTipType, x, y)
    self.wndStrict:SetTooltip(
        (self.app.client.course.strictMode and
         "Set the course so that, when racing, the waypoints can be visited in any order") or
         "Set the course so that, when racing, the waypoints must be visited in the exact order specified"
    )
end

function ParaseleneWildRace.Editor:SetCourseAndIndex(course, index)
    local scrollPosition = self.app.client:GetScrollPosition()
    self.app.client:SetCourse(course)
    self.app.client:SetIndex(index)
    self.app.client:SetScrollPosition(scrollPosition)
    self.app.client:AdjustScrollbar()
end

function ParaseleneWildRace.Editor:SetWindowLock(isLocked)
    if isLocked then
        self.app.characterData.editorWindowInfo.locked = true
        self.wndEditor:SetStyle("Moveable", false)
        self.wndLockIcon:SetSprite(ParaseleneWildRace.lockEnabledIcon)
    else
        self.app.characterData.editorWindowInfo.locked = false
        self.wndEditor:SetStyle("Moveable", true)
        self.wndLockIcon:SetSprite(ParaseleneWildRace.lockDisabledIcon)
    end
end

function ParaseleneWildRace.Editor:ShowWindow()
    if not self.wndEditor:IsVisible() then
        self:EnableEditboxes()
        self.wndEditor:Invoke()
        self.app.client.wndEditorFrame:SetSprite(
            (self.app.client.btnEditor:ContainsMouse() and ParaseleneWildRace.pressedToolbarButtonFlybyFrame) or
            ParaseleneWildRace.pressedToolbarButtonFrame
        )
    end
end

function ParaseleneWildRace.Editor:ToggleWindowVisibility()
    if self.wndEditor:IsVisible() then
        self:HideWindow()
    else
        self:ShowWindow()
    end
end
