-- Paraselene's WildRace
-- Version 0.5.2

require "Apollo"
require "Window"
 
ParaseleneWildRace.Files = {}

function ParaseleneWildRace.Files:new(app)
    local i = {}
    setmetatable(i, self)
    self.__index = self

    i.app = app

    if not i.app.characterData.filesWindowInfo.width or
       (i.app.characterData.filesWindowInfo.width < ParaseleneWildRace.minFilesWindowWidth) then
        i.app.characterData.filesWindowInfo.width = ParaseleneWildRace.minFilesWindowWidth
    end
    if not i.app.characterData.filesWindowInfo.height or
       (i.app.characterData.filesWindowInfo.height < ParaseleneWildRace.minFilesWindowHeight) then
        i.app.characterData.filesWindowInfo.height = ParaseleneWildRace.minFilesWindowHeight
    end

    i.wndFiles = Apollo.LoadForm(i.app.doc, "wndFiles", nil, i)
    if not i.wndFiles then
        local message = ParaseleneWildRace.appName ..
            ": Error: Could not load the \"Files\" window."
        ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_System, message)
        Apollo.AddAddonErrorText(ParaseleneWildRaceInstance, message)
        return nil
    end

    local wndTitle = i.wndFiles:FindChild("wndTitle")
    i.btnAbout = wndTitle:FindChild("btnAbout")
    i.btnLock = wndTitle:FindChild("btnLock")
    i.wndLockIcon = i.btnLock:FindChild("wndIcon")

    i.wndTreeList = i.wndFiles:FindChild("wndTree"):FindChild("wndTreeList")

    local wndTools = i.wndFiles:FindChild("wndTools")

    local wndCourseTools = wndTools:FindChild("wndCourseTools")
    i.btnLoad = wndCourseTools:FindChild("btnLoad")
    i.btnLoad:SetData(ParaseleneWildRace.Button:new{
        button = i.btnLoad,
        style = ParaseleneWildRace.buttonStyles.red
    })
    i.btnSave = wndCourseTools:FindChild("btnSave")
    i.btnSave:SetData(ParaseleneWildRace.Button:new{
        button = i.btnSave
    })

    local wndFolderTools = wndTools:FindChild("wndFolderTools")
    i.btnCreate = wndFolderTools:FindChild("btnCreate")
    i.btnCreate:SetData(ParaseleneWildRace.Button:new{
        button = i.btnCreate
    })

    local wndCommonTools = wndTools:FindChild("wndCommonTools")
    i.btnMoveUp = wndCommonTools:FindChild("btnMoveUp")
    i.btnMoveUp:SetData(ParaseleneWildRace.Button:new{
        button = i.btnMoveUp
    })
    i.btnMoveDown = wndCommonTools:FindChild("btnMoveDown")
    i.btnMoveDown:SetData(ParaseleneWildRace.Button:new{
        button = i.btnMoveDown
    })
    i.btnCopy = wndCommonTools:FindChild("btnCopy")
    i.btnCopy:SetData(ParaseleneWildRace.Button:new{
        button = i.btnCopy
    })
    i.btnPaste = wndCommonTools:FindChild("btnPaste")
    i.btnPaste:SetData(ParaseleneWildRace.Button:new{
        button = i.btnPaste
    })
    i.btnDelete = wndCommonTools:FindChild("btnDelete")
    i.btnDelete:SetData(ParaseleneWildRace.Button:new{
        button = i.btnDelete,
        style = ParaseleneWildRace.buttonStyles.red
    })
    i.wndRename = wndCommonTools:FindChild("wndRename")
    i.edtRename = i.wndRename:FindChild("wndRename"):FindChild("edtRename")

    i.indexToItemMap = {}

    wndTitle:SetText(ParaseleneWildRace.appName .. " Files")

    i.wndFiles:SetSizingMinimum(ParaseleneWildRace.minFilesWindowWidth, ParaseleneWildRace.minFilesWindowHeight)
    i.wndFiles:Move(
        i.app.characterData.filesWindowInfo.left,
        i.app.characterData.filesWindowInfo.top,
        i.app.characterData.filesWindowInfo.width,
        i.app.characterData.filesWindowInfo.height
    )
    i:SetWindowLock(i.app.characterData.filesWindowInfo.locked)

    i:SetStorage(nil)

    i.wndFiles:SetSprite(ParaseleneWildRace.backgroundSprites[i.app.characterData.uiOptions.background])
    i.wndFiles:Show(false, true)

    return i
end

function ParaseleneWildRace.Files:AdjustScrollbar()
    if self.index then
        local windowHeight = self.wndTreeList:GetHeight()
        local listHeight = #self.indexToItemMap * ParaseleneWildRace.treeListItemHeight
        local itemOffset = (self.index - 1) * ParaseleneWildRace.treeListItemHeight
        local currentScrollPosition = self.wndTreeList:GetVScrollPos()
        local newScrollPosition = currentScrollPosition
        if self.index > 2 then
            if newScrollPosition > (itemOffset - ParaseleneWildRace.treeListItemHeight) then
                newScrollPosition = itemOffset - ParaseleneWildRace.treeListItemHeight
            end
        else
            newScrollPosition = 0
        end
        if self.index < #self.indexToItemMap then
            if (newScrollPosition + windowHeight) < (itemOffset + (2 * ParaseleneWildRace.treeListItemHeight)) then
                newScrollPosition = itemOffset + (2 * ParaseleneWildRace.treeListItemHeight) - windowHeight
            end
        else
            if (newScrollPosition + windowHeight) < (itemOffset + ParaseleneWildRace.treeListItemHeight) then
                newScrollPosition = itemOffset + ParaseleneWildRace.treeListItemHeight - windowHeight
            end
        end
        if newScrollPosition ~= currentScrollPosition then
            self.wndTreeList:SetVScrollPos(newScrollPosition)
        end
    end
end

function ParaseleneWildRace.Files:AdjustTools()
    local item = self.index and self.indexToItemMap[self.index]
    local siblings, siblingIndex
    if self.storage and item then
        local _, _, s, i = self.storage:GetFamily(ParaseleneWildRace.Storage.GetPath(item))
        siblings = s
        siblingIndex = i
    end

    -- Enable or disable the load button.
    if item and (item.folderItemType == ParaseleneWildRace.folderItemTypes.course) then
        self.btnLoad:GetData():Enable()
    else
        self.btnLoad:GetData():Disable()
    end

    -- Enable or disable the save button.
    if item and self.app.client and self.app.client.course then
        self.btnSave:GetData():Enable()
    else
        self.btnSave:GetData():Disable()
    end

    -- Enable or disable the create folder button.
    if item then
        self.btnCreate:GetData():Enable()
    else
        self.btnCreate:GetData():Disable()
    end

    -- Enable or disable the move up button.
    if siblingIndex and (siblingIndex > 1) then
        self.btnMoveUp:GetData():Enable()
    else
        self.btnMoveUp:GetData():Disable()
    end

    -- Enable or disable the move down button.
    if siblingIndex and (siblingIndex < #siblings) then
        self.btnMoveDown:GetData():Enable()
    else
        self.btnMoveDown:GetData():Disable()
    end

    -- Enable or disable the copy button.
    if item then
        self.btnCopy:GetData():Enable()
    else
        self.btnCopy:GetData():Disable()
    end

    -- Enable or disable the paste button.
    if item and self.clipboard then
        self.btnPaste:GetData():Enable()
    else
        self.btnPaste:GetData():Disable()
    end

    -- Enable or disable the delete button.
    if item and (item.uid ~= self.storage.baseFolder.uid) then
        self.btnDelete:GetData():Enable()
    else
        self.btnDelete:GetData():Disable()
    end

    self.isRenameEnabled = item and (item.uid ~= self.storage.baseFolder.uid)

    -- Color the rename label and editbox as either enabled or disabled.
    if self.isRenameEnabled then
        self.wndRename:SetTextColor(ApolloColor.new(ParaseleneWildRace.enabledTextColor))
        self.edtRename:SetTextColor(ApolloColor.new(ParaseleneWildRace.enabledTextColor))
    else
        self.wndRename:SetTextColor(ApolloColor.new(ParaseleneWildRace.disabledTextColor))
        self.edtRename:SetTextColor(ApolloColor.new(ParaseleneWildRace.disabledTextColor))
    end

    -- Enable or disable the rename editbox.
    if self.isRenameEnabled then
        self.edtRename:SetStyle("IgnoreMouse", false)
        self.edtRename:SetStyleEx("ReadOnly", false)
    else
        self.edtRename:SetStyle("IgnoreMouse", true)
        self.edtRename:SetStyleEx("ReadOnly", true)
        self.edtRename:ClearFocus() -- Keep the rename from unwanted highlighted.
    end

    -- Fill the rename edit box with the name of the currently indexed item.
    self.edtRename:SetText((item and item.name) or "")
end

function ParaseleneWildRace.Files:BuildTree(folder, depth, startIndex)
    local index = startIndex

    folder.wndTreeListItem = Apollo.LoadForm(self.app.doc, "wndTreeListItem", self.wndTreeList, self)
    if not folder.wndTreeListItem then
        return nil
    end

    local wndContent = folder.wndTreeListItem:FindChild("wndContent")
    wndContent:SetAnchorOffsets(depth * ParaseleneWildRace.treeListItemIndent, 0, 0, 0) -- Indent.

    local wndType = wndContent:FindChild("wndType")
    wndType:SetSprite(ParaseleneWildRace.treeFolderIcon)
    wndType:SetTooltip("Folder")

    local edtName = wndContent:FindChild("edtName")
    edtName:SetText(folder.name)
    edtName:SetTooltip(folder.name)

    folder.wndTreeListItem:SetData(index)
    self.indexToItemMap[index] = folder
    folder.index = index
    index = index + 1

    local btnExpandCollapse = wndContent:FindChild("btnExpandCollapse")
    if folder.collapsed then
        btnExpandCollapse:FindChild("wndIcon"):SetSprite(ParaseleneWildRace.treeExpandFolderIcon)
        btnExpandCollapse:SetTooltip("Expand folder \"" .. folder.name .. "\"")
    else
        btnExpandCollapse:FindChild("wndIcon"):SetSprite(ParaseleneWildRace.treeCollapseFolderIcon)
        btnExpandCollapse:SetTooltip("Collapse folder \"" .. folder.name .. "\"")

        for idx, childFolder in ipairs(folder.folders) do
            index = self:BuildTree(childFolder, depth + 1, index)
            if not index then
                return nil
            end
        end

        for idx, course in ipairs(folder.courses) do
            course.wndTreeListItem = Apollo.LoadForm(self.app.doc, "wndTreeListItem", self.wndTreeList, self)
            if not course.wndTreeListItem then
                return nil
            end

            wndContent = course.wndTreeListItem:FindChild("wndContent")
            wndContent:SetAnchorOffsets((depth + 1) * ParaseleneWildRace.treeListItemIndent, 0, 0, 0) -- Indent.

            wndType = wndContent:FindChild("wndType")
            wndType:SetSprite(ParaseleneWildRace.treeCourseIcon)
            wndType:SetTooltip("Course")

            edtName = wndContent:FindChild("edtName")
            edtName:SetText(course.name)
            edtName:SetTooltip(course.name)

            course.wndTreeListItem:SetData(index)
            self.indexToItemMap[index] = course
            course.index = index
            index = index + 1

            wndContent:FindChild("btnExpandCollapse"):FindChild("wndIcon"):SetSprite("")
        end

    end

    return index
end

function ParaseleneWildRace.Files:ClearStorage()
    if self.storage then
        for idx, item in ipairs(self.indexToItemMap) do
            if item.wndTreeListItem then
                item.wndTreeListItem:Destroy()
                item.wndTreeListItem = nil
            end
        end
        self.storage = nil
        self.indexToItemMap = {}

        -- index is set to nil here in order to maintain integrity between storage and index.
        -- SetIndex() expects self.index to be nil if self.indexToItemMap is empty.
        self.index = nil
    end
    self:SetIndex(nil)
end

--[[
function ParaseleneWildRace.Files.DeepCompareArray(array1, array2)
    if #array1 ~= #array2 then
        return false
    end
    for idx = 1, #array1 do
        if type(array1[idx]) ~= type(array2[idx]) then
            return false
        elseif type(array1[idx]) == "table" then
            if not ParaseleneWildRace.Files.DeepCompareArray(array1[idx], array2[idx]) then
                return false
            end
        elseif array1[idx] ~= array2[idx] then
            return false
        end
    end
    return true
end
--]]

function ParaseleneWildRace.Files.DeepCopyItem(item)
    if item.folderItemType == ParaseleneWildRace.folderItemTypes.folder then
        return ParaseleneWildRace.Folder:new(item)
    else -- item.folderItemType == ParaseleneWildRace.folderItemTypes.course
        return ParaseleneWildRace.Course:new(item)
    end
end

function ParaseleneWildRace.Files:DisableEditboxes()
    -- Reset the rename editbox to clear out any unfinished editing.
    local item = self.index and self.indexToItemMap[self.index]
    local name = (item and item.name) or ""
    self.edtRename:SetText(name)
    self.edtRename:ClearFocus() -- Keep the rename editbox from unwanted highlighted.
end

function ParaseleneWildRace.Files:GetScrollPosition()
    return self.wndTreeList:GetVScrollPos()
end

function ParaseleneWildRace.Files:GetStorage()
    local storage = ParaseleneWildRace.Storage:new(self.storage)
    local path = ParaseleneWildRace.Storage.GetPath(self.indexToItemMap[self.index])
    return storage, path
end

function ParaseleneWildRace.Files:HideWindow()
    if self.wndFiles:IsVisible() then
        self:DisableEditboxes()
        self.wndFiles:Close()
        self.app.client.wndFilesFrame:SetSprite(
            (self.app.client.btnFiles:ContainsMouse() and ParaseleneWildRace.enabledToolbarButtonFlybyFrame) or
            ParaseleneWildRace.enabledToolbarButtonFrame
        )
    end
end

function ParaseleneWildRace.Files:OnBtnAbout(wndHandler, wndControl, eMouseButton)
    self.app.about:ToggleWindowVisibility()
    Sound.Play(ParaseleneWildRace.buttonClickSound)
end

function ParaseleneWildRace.Files:OnBtnAboutGenerateTooltip(wndHandler, wndControl, eToolTipType, x, y)
    self.btnAbout:SetTooltip(
        (self.app.about.wndAbout:IsVisible() and "Close the \"About\" window") or
        "Open the \"About\" window"
    )
end

function ParaseleneWildRace.Files:OnBtnClose(wndHandler, wndControl, eMouseButton)
    self:HideWindow()
    Sound.Play(ParaseleneWildRace.buttonClickSound)
end

function ParaseleneWildRace.Files:OnBtnCopy(wndHandler, wndControl, eMouseButton)
    if self.btnCopy:GetData():IsEnabled() then
        self.clipboard = ParaseleneWildRace.Files.DeepCopyItem(self.indexToItemMap[self.index])
        self:AdjustTools()
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    end
end

function ParaseleneWildRace.Files:OnBtnCopyGenerateTooltip(wndHandler, wndControl, eToolTipType, x, y)
    local item = self.index and self.indexToItemMap[self.index]
    if item then
        local itemType = ((item.folderItemType == ParaseleneWildRace.folderItemTypes.folder) and "folder") or "course"
        self.btnCopy:SetTooltip("Copy the selected " .. itemType .. " to the clipboard")
    else
        self.btnCopy:SetTooltip("Copy the selected folder or course to the clipboard")
    end
end

function ParaseleneWildRace.Files:OnBtnCreate(wndHandler, wndControl, eMouseButton)
    if self.btnCreate:GetData():IsEnabled() then
        local storage, path = self:GetStorage()
        local item, parent = storage:GetFamily(path)
        local newFolder = ParaseleneWildRace.Folder:new{
            name = "New Folder"
        }
        if item.folderItemType == ParaseleneWildRace.folderItemTypes.folder then
            item:Append(newFolder)
            table.insert(path, newFolder.uid)
            item.collapsed = false -- Expand the parent directory.
        else -- item.folderItemType == ParaseleneWildRace.folderItemTypes.course
            parent:Append(newFolder)
            path[#path] = newFolder.uid
        end
        self:SetStorageAndIndex(storage, path) -- Keep the selection on the new folder.
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    end
end

function ParaseleneWildRace.Files:OnBtnDelete(wndHandler, wndControl, eMouseButton)
    if self.btnDelete:GetData():IsEnabled() then
        local storage, path = self:GetStorage()
        local item, parent, siblings, index = storage:GetFamily(path)
        parent:Delete(index, item)
        if #siblings == 0 then -- If we deleted the only item in the folder ...
            table.remove(path, #path) -- ... then point the path to the parent.
        elseif index > #siblings then -- If we deleted the tail item in the folder ...
            path[#path] = siblings[#siblings].uid -- ... then point the path to the new tail item in the folder.
        else
            path[#path] = siblings[index].uid -- ... Point the path to the next item in the folder.
        end
        self:SetStorageAndIndex(storage, path)
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    end
end

function ParaseleneWildRace.Files:OnBtnDeleteGenerateTooltip(wndHandler, wndControl, eToolTipType, x, y)
    local item = self.index and self.indexToItemMap[self.index]
    if item then
        local itemType = ((item.folderItemType == ParaseleneWildRace.folderItemTypes.folder) and "folder") or "course"
        self.btnDelete:SetTooltip("Delete the selected " .. itemType)
    else
        self.btnDelete:SetTooltip("Delete the selected folder or course")
    end
end

function ParaseleneWildRace.Files:OnBtnExpandCollapse(wndHandler, wndControl, eMouseButton)
    local index = wndHandler:GetParent():GetParent():GetData() -- index of wndTreeListItem.
    local folder = self.indexToItemMap[index]
    folder.collapsed = not folder.collapsed
    self:SetStorageAndIndex(self.storage, ParaseleneWildRace.Storage.GetPath(folder))
    Sound.Play(ParaseleneWildRace.buttonClickSound)
end

function ParaseleneWildRace.Files:OnBtnLoad(wndHandler, wndControl, eMouseButton)
    if self.btnLoad:GetData():IsEnabled() then
        self.app.client:SetCourse(self.indexToItemMap[self.index])
        if self.app.client.course then -- Only if SetCourse() succeeded. It can fail if we have a LoadForm() failure.
            self.app.client.course.uid = 0 -- Just to be clean.
        end
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    end
end

function ParaseleneWildRace.Files:OnBtnLock(wndHandler, wndControl, eMouseButton)
    self:SetWindowLock(not self.app.characterData.filesWindowInfo.locked)
    Sound.Play(ParaseleneWildRace.buttonClickSound)
end

function ParaseleneWildRace.Files:OnBtnLockGenerateTooltip(wndHandler, wndControl, eToolTipType, x, y)
    self.btnLock:SetTooltip(
        (self.app.characterData.filesWindowInfo.locked and "Unlock the window so that it can be moved and resized") or
        "Lock the window so that it cannot be moved or resized"
    )
end

function ParaseleneWildRace.Files:OnBtnMoveDown(wndHandler, wndControl, eMouseButton)
    if self.btnMoveDown:GetData():IsEnabled() then
        local storage, path = self:GetStorage()
        storage:ShiftLater(path)
        self:SetStorageAndIndex(storage, path) -- Keep the selection on the moving item.
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    end
end

function ParaseleneWildRace.Files:OnBtnMoveDownGenerateTooltip(wndHandler, wndControl, eToolTipType, x, y)
    local item = self.index and self.indexToItemMap[self.index]
    if item then
        local itemType = ((item.folderItemType == ParaseleneWildRace.folderItemTypes.folder) and "folder") or "course"
        self.btnMoveDown:SetTooltip("Move the selected " .. itemType .. " downward in its parent folder")
    else
        self.btnMoveDown:SetTooltip("Move the selected folder or course downward in its parent folder")
    end
end

function ParaseleneWildRace.Files:OnBtnMoveUp(wndHandler, wndControl, eMouseButton)
    if self.btnMoveUp:GetData():IsEnabled() then
        local storage, path = self:GetStorage()
        storage:ShiftEarlier(path)
        self:SetStorageAndIndex(storage, path) -- Keep the selection on the moving item.
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    end
end

function ParaseleneWildRace.Files:OnBtnMoveUpGenerateTooltip(wndHandler, wndControl, eToolTipType, x, y)
    local item = self.index and self.indexToItemMap[self.index]
    if item then
        local itemType = ((item.folderItemType == ParaseleneWildRace.folderItemTypes.folder) and "folder") or "course"
        self.btnMoveUp:SetTooltip("Move the selected " .. itemType .. " upward in its parent folder")
    else
        self.btnMoveUp:SetTooltip("Move the selected folder or course upward in its parent folder")
    end
end

function ParaseleneWildRace.Files:OnBtnPaste(wndHandler, wndControl, eMouseButton)
    if self.btnPaste:GetData():IsEnabled() then
        local storage, path = self:GetStorage()
        local item, parent, _, index = storage:GetFamily(path)
        local newItem = ParaseleneWildRace.Files.DeepCopyItem(self.clipboard)
        if item.folderItemType == ParaseleneWildRace.folderItemTypes.folder then
            item:Append(newItem)
            table.insert(path, newItem.uid)
            item.collapsed = false -- Expand the parent directory.
        else -- item.folderItemType == ParaseleneWildRace.folderItemTypes.course
            if newItem.folderItemType == ParaseleneWildRace.folderItemTypes.folder then
                parent:Append(newItem)
            else -- newItem.folderItemType == ParaseleneWildRace.folderItemTypes.course
                parent:Insert(index, newItem)
            end
            path[#path] = newItem.uid
        end
        self:SetStorageAndIndex(storage, path) -- Keep the selection on the pasted waypoint.
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    end
end

function ParaseleneWildRace.Files:OnBtnPasteGenerateTooltip(wndHandler, wndControl, eToolTipType, x, y)
    local srcType = (
        (self.clipboard and
            (((self.clipboard.folderItemType == ParaseleneWildRace.folderItemTypes.folder) and
                "folder") or "course"
            )
        ) or "folder or course"
    )
    local dstItem = self.index and self.indexToItemMap[self.index]
    if dstItem then
        if self.indexToItemMap[self.index].folderItemType == ParaseleneWildRace.folderItemTypes.folder then
            self.btnPaste:SetTooltip("Paste the " .. srcType .. " in the clipboard into the selected folder")
        else -- self.indexToItemMap[self.index].folderItemType == ParaseleneWildRace.folderItemTypes.course
            self.btnPaste:SetTooltip(
                "Paste the " .. srcType .. " in the clipboard into the parent folder of the selected course"
            )
        end
    else
        self.btnPaste:SetTooltip(
            "Paste the " .. srcType .. " in the clipboard into the selected folder " ..
            "or the parent folder of the selected course"
        )
    end
end

function ParaseleneWildRace.Files:OnBtnSave(wndHandler, wndControl, eMouseButton)
    if self.btnSave:GetData():IsEnabled() then
        local storage, path = self:GetStorage()
        local item, parent, _, index = storage:GetFamily(path)
        local course = ParaseleneWildRace.Files.DeepCopyItem(self.app.client.course)
        if item.folderItemType == ParaseleneWildRace.folderItemTypes.folder then
            item:Append(course)
            table.insert(path, course.uid)
            item.collapsed = false -- Expand the parent directory.
        else -- item.folderItemType == ParaseleneWildRace.folderItemTypes.course
            parent:Insert(index, course)
            path[#path] = course.uid
        end
        self:SetStorageAndIndex(storage, path) -- Keep the selection on the saved course.
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    end
end

function ParaseleneWildRace.Files:OnButtonMouseEnter(wndHandler, wndControl, x, y)
    wndHandler:GetData():OnMouseEnter()
end

function ParaseleneWildRace.Files:OnButtonMouseExit(wndHandler, wndControl, x, y)
    wndHandler:GetData():OnMouseExit()
end

function ParaseleneWildRace.Files:OnEdtRenameEscape(wndHandler, wndControl)
    -- Reset the item name to clear out any unfinished editing.
    self.edtRename:SetText(self.indexToItemMap[self.index].name)
end

function ParaseleneWildRace.Files:OnEdtRenameReturn(wndHandler, wndControl, strText)
    local newText = string.gsub(strText, "^%s*(.-)%s*$", "%1") -- Strip leading and trailing spaces.
    if newText ~= "" then
        local storage, path = self:GetStorage()
        local item = storage:GetItem(path)
        item.name = newText
        self:SetStorageAndIndex(storage, path) -- Preserve the selection.
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    else
        self:OnEdtRenameEscape(wndHandler, wndControl)
    end
end

function ParaseleneWildRace.Files:OnItemSelected(
        wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY, bDoubleClick, bStopPropagation)
    local index = wndHandler:GetData()
    if index and index ~= self.index then
        self:SetIndex(index)
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    end
end

function ParaseleneWildRace.Files:OnWindowMove(wndHandler, wndControl, nOldLeft, nOldTop, nOldRight, nOldBottom)
    self:OnWindowMoveOrSizeChanged()
end

function ParaseleneWildRace.Files:OnWindowMoveOrSizeChanged()
    -- Keep self.app.characterData.filesWindowInfo up to date with the window's position and size.
    local left, top, right, bottom = self.wndFiles:GetRect()
    self.app.characterData.filesWindowInfo.left = left or 0
    self.app.characterData.filesWindowInfo.top = top or 0
    self.app.characterData.filesWindowInfo.width = (right and (right - left)) or ParaseleneWildRace.minFilesWindowWidth
    self.app.characterData.filesWindowInfo.height = (bottom and (bottom - top)) or ParaseleneWildRace.minFilesWindowHeight
end

function ParaseleneWildRace.Files:OnWindowSizeChanged(wndHandler, wndControl)
    self:OnWindowMoveOrSizeChanged()
end

function ParaseleneWildRace.Files:OnWndRenameGenerateTooltip(wndHandler, wndControl, eToolTipType, x, y)
    local item = self.index and self.indexToItemMap[self.index]
    if item then
        local itemType = ((item.folderItemType == ParaseleneWildRace.folderItemTypes.folder) and "folder") or "course"
        self.wndRename:SetTooltip("Rename the selected " .. itemType)
    else
        self.wndRename:SetTooltip("Rename the selected folder or course")
    end
end

function ParaseleneWildRace.Files:SetIndex(index)
    -- Remove the border around the currently selected item in the tree.
    if self.index then
        self.indexToItemMap[self.index].wndTreeListItem:SetSprite("")
    end

    if self.storage then
        self.index = (index and self.indexToItemMap[index] and index) or 1

        -- Set the border around the newly selected item in the tree.
        local wndContent = self.indexToItemMap[self.index].wndTreeListItem:FindChild("wndContent")
        self.indexToItemMap[self.index].wndTreeListItem:SetSprite(ParaseleneWildRace.selectedItemBorder)
    else
        self.index = nil
    end

    self:AdjustTools()

    -- Keep the rename editbox from unwanted highlighted.
    -- Unwanted highlighting occurs when text is set in an editbox that was once editable (non-ReadOnly).
    self.edtRename:ClearFocus()
end

function ParaseleneWildRace.Files:SetScrollPosition(position)
    local windowHeight = self.wndTreeList:GetHeight()
    local listHeight = #self.indexToItemMap * ParaseleneWildRace.treeListItemHeight
    if (position + windowHeight) > listHeight then
        position = listHeight - windowHeight
    end
    if position < 0 then
        position = 0
    end
    if position ~= self.wndTreeList:GetVScrollPos() then
        self.wndTreeList:SetVScrollPos(position)
    end
end

function ParaseleneWildRace.Files:SetStorage(storage)
    self:ClearStorage()
    if storage then
        self.storage = ParaseleneWildRace.Storage:new(storage) -- Deep copy.

        if not self:BuildTree(self.storage.baseFolder, 0, 1) then
            local message = ParaseleneWildRace.appName ..
                ": Error: Could not load a tree list item window."
            ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_System, message)
            Apollo.AddAddonErrorText(ParaseleneWildRaceInstance, message)

            self:ClearStorage()
            return
        end
        self.wndTreeList:ArrangeChildrenVert()
        self:SetIndex(1)
    end
end

function ParaseleneWildRace.Files:SetStorageAndIndex(storage, path)
    local scrollPosition = self:GetScrollPosition()
    self:SetStorage(storage)
    self:SetIndex(self.storage:GetItem(path).index)
    self:SetScrollPosition(scrollPosition)
    self:AdjustScrollbar()
end

function ParaseleneWildRace.Files:SetWindowLock(isLocked)
    if isLocked then
        self.app.characterData.filesWindowInfo.locked = true
        self.wndFiles:SetStyle("Moveable", false)
        self.wndFiles:SetStyle("Sizable", false)
        self.wndLockIcon:SetSprite(ParaseleneWildRace.lockEnabledIcon)
    else
        self.app.characterData.filesWindowInfo.locked = false
        self.wndFiles:SetStyle("Moveable", true)
        self.wndFiles:SetStyle("Sizable", true)
        self.wndLockIcon:SetSprite(ParaseleneWildRace.lockDisabledIcon)
    end
end

function ParaseleneWildRace.Files:ShowWindow()
    if not self.wndFiles:IsVisible() then
        self.wndFiles:Invoke()
        self.app.client.wndFilesFrame:SetSprite(
            (self.app.client.btnFiles:ContainsMouse() and ParaseleneWildRace.pressedToolbarButtonFlybyFrame) or
            ParaseleneWildRace.pressedToolbarButtonFrame
        )
    end
end

function ParaseleneWildRace.Files:ToggleWindowVisibility()
    if self.wndFiles:IsVisible() then
        self:HideWindow()
    else
        self:ShowWindow()
    end
end
