-- Paraselene's WildRace
-- Version 0.5.2

ParaseleneWildRace.Folder = {}
 
function ParaseleneWildRace.Folder:new(o)
    if type(o) ~= "table" then
        return nil
    end

    local i = {}
    setmetatable(i, self)
    self.__index = self 

    i.name = ((type(o.name) == "string") and o.name) or nil
    i.uid = ((type(o.uid) == "number") and o.uid) or 0 -- Zero means "unassigned" to a folder.

    -- Does not vary, so, no need for check or persistence.
    i.folderItemType = ParaseleneWildRace.folderItemTypes.folder

    i.collapsed = (type(o.collapsed) == "boolean") and o.collapsed

    -- i.folders cannot be nil, even if the o.folders is nil.
    -- i.courses cannot be nil, even if the o.courses is nil.
    -- Note: These are defined before the following loops because GetNewChildUID() needs them to exist.
    i.folders = {}
    i.courses = {}

    if o.folders and (type(o.folders) == "table") then
        for idx, folder in ipairs(o.folders) do
            i.folders[idx] = ParaseleneWildRace.Folder:new(folder)
            if not i.folders[idx] then
                return nil
            end
            i.folders[idx].parent = i
            if i.folders[idx].uid == 0 then
                i.folders[idx].uid = i:GetNewChildUID()
            end
        end
    end

    if o.courses and (type(o.courses) == "table") then
        for idx, course in ipairs(o.courses) do
            i.courses[idx] = ParaseleneWildRace.Course:new(course)
            if not i.courses[idx] then
                return nil
            end
            i.courses[idx].parent = i
            if i.courses[idx].uid == 0 then
                i.courses[idx].uid = i:GetNewChildUID()
            end
        end
    end

    return (i:IsValid() and i) or nil
end

function ParaseleneWildRace.Folder:Append(item)
    local newChildUID = self:GetNewChildUID()
    local children
    if item.folderItemType == ParaseleneWildRace.folderItemTypes.folder then
        children = self.folders
    else -- item.folderItemType == ParaseleneWildRace.folderItemTypes.course
        children = self.courses
    end
    table.insert(children, item)
    item.parent = self
    item.uid = newChildUID
end

function ParaseleneWildRace.Folder:Delete(index, item)
    local children
    if item.folderItemType == ParaseleneWildRace.folderItemTypes.folder then
        children = self.folders
    else -- item.folderItemType == ParaseleneWildRace.folderItemTypes.course
        children = self.courses
    end
    table.remove(children, index)
end

function ParaseleneWildRace.Folder:GetNewChildUID()
    local existingUIDs = {}

    for idx, folder in ipairs(self.folders) do
        existingUIDs[folder.uid] = true
    end

    for idx, course in ipairs(self.courses) do
        existingUIDs[course.uid] = true
    end

    local newUID = 1 -- Zero means "unassigned" to a folder. One and above are "assigned" UIDs.
    while existingUIDs[newUID] do
        newUID = newUID + 1
    end

    return newUID
end

function ParaseleneWildRace.Folder:Insert(index, item)
    local newChildUID = self:GetNewChildUID()
    local children
    if item.folderItemType == ParaseleneWildRace.folderItemTypes.folder then
        children = self.folders
    else -- item.folderItemType == ParaseleneWildRace.folderItemTypes.course
        children = self.courses
    end
    table.insert(children, index, item)
    item.parent = self
    item.uid = newChildUID
end

function ParaseleneWildRace.Folder:IsValid()
    -- self.uid does not need to be checked. It will always exist.
    -- self.collapsed does not need to be checked. It will always exist.
    -- self.folders does not need to be checked. It will always exist.
    -- self.courses does not need to be checked. It will always exist.
    return (self.name ~= nil)
end

function ParaseleneWildRace.Folder:Save()
    local folders = {}
    for idx, folder in ipairs(self.folders) do
        folders[idx] = folder:Save()
    end

    local courses = {}
    for idx, course in ipairs(self.courses) do
        courses[idx] = course:Save()
    end

    return {
        name = self.name,
        uid = self.uid,
        collapsed = self.collapsed,
        folders = folders,
        courses = courses
    }
end
