-- Paraselene's WildRace
-- Version 0.5.2

require "Apollo"
require "Window"
 
ParaseleneWildRace.Options = {}

function ParaseleneWildRace.Options:new(app)
    local i = {}
    setmetatable(i, self)
    self.__index = self

    i.app = app

    i.wndOptions = Apollo.LoadForm(i.app.doc, "wndOptions", nil, i)
    if not i.wndOptions then
        local message = ParaseleneWildRace.appName ..
            ": Error: Could not load the \"Options\" window."
        ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_System, message)
        Apollo.AddAddonErrorText(ParaseleneWildRaceInstance, message)
        return nil
    end

    local wndTitle = i.wndOptions:FindChild("wndTitle")
    i.btnAbout = wndTitle:FindChild("btnAbout")
    i.btnLock = wndTitle:FindChild("btnLock")
    i.wndLockIcon = i.btnLock:FindChild("wndIcon")

    local wndTools = i.wndOptions:FindChild("wndTools")

    local wndUserInterface = wndTools:FindChild("wndUserInterface")
    local wndBackgroundGroup = wndUserInterface:FindChild("wndBackground"):FindChild("wndGroup")
    i.wndHolo = wndBackgroundGroup:FindChild("wndHolo")
    i.btnHolo = i.wndHolo:FindChild("btnHolo")
    i.wndHoloIcon = i.btnHolo:FindChild("wndIcon")
    i.wndCircuit = wndBackgroundGroup:FindChild("wndCircuit")
    i.btnCircuit = i.wndCircuit:FindChild("btnCircuit")
    i.wndCircuitIcon = i.btnCircuit:FindChild("wndIcon")

    local wndVisibleWaypoints = wndTools:FindChild("wndVisibleWaypoints")
    i.wndTrackMaster = wndVisibleWaypoints:FindChild("wndTrackMaster")
    i.btnTrackMaster = i.wndTrackMaster:FindChild("btnTrackMaster")
    i.wndTrackMasterIcon = i.btnTrackMaster:FindChild("wndIcon")

    wndTitle:SetText(ParaseleneWildRace.appName .. " Options")

    local width = ParaseleneWildRace.optionsWindowWidth
    local height = ParaseleneWildRace.optionsWindowHeight
    i.wndOptions:Move(
        i.app.characterData.optionsWindowInfo.left,
        i.app.characterData.optionsWindowInfo.top,
        width,
        height
    )
    i:SetWindowLock(i.app.characterData.optionsWindowInfo.locked)

    i.wndOptions:SetSprite(ParaseleneWildRace.backgroundSprites[i.app.characterData.uiOptions.background])
    i.wndOptions:Show(false, true)

    return i
end

function ParaseleneWildRace.Options:AdjustOptions()
    self:AdjustUserInterfaceOptions()
    self:AdjustVisibleWaypointsOptions()
end

function ParaseleneWildRace.Options:AdjustUserInterfaceOptions()
    -- Select or unselect the holo radio button.
    if self.app.characterData.uiOptions and
       (self.app.characterData.uiOptions.background == ParaseleneWildRace.backgrounds.holo) then
        self.btnHolo:SetData(true) -- Checked. Used by mouse enter/exit handlers.
        self.wndHoloIcon:SetSprite(
            (self.btnHolo:ContainsMouse() and ParaseleneWildRace.checkedRadioFlybyIcon) or
            ParaseleneWildRace.checkedRadioIcon
        )
    else
        self.btnHolo:SetData(false) -- Unchecked. Used by mouse enter/exit handlers.
        self.wndHoloIcon:SetSprite(
            (self.btnHolo:ContainsMouse() and ParaseleneWildRace.uncheckedRadioFlybyIcon) or
            ParaseleneWildRace.uncheckedRadioIcon
        )
    end

    -- Select or unselect the circuit radio button.
    if self.app.characterData.uiOptions and
       (self.app.characterData.uiOptions.background == ParaseleneWildRace.backgrounds.circuit) then
        self.btnCircuit:SetData(true) -- Checked. Used by mouse enter/exit handlers.
        self.wndCircuitIcon:SetSprite(
            (self.btnCircuit:ContainsMouse() and ParaseleneWildRace.checkedRadioFlybyIcon) or
            ParaseleneWildRace.checkedRadioIcon
        )
    else
        self.btnCircuit:SetData(false) -- Unchecked. Used by mouse enter/exit handlers.
        self.wndCircuitIcon:SetSprite(
            (self.btnCircuit:ContainsMouse() and ParaseleneWildRace.uncheckedRadioFlybyIcon) or
            ParaseleneWildRace.uncheckedRadioIcon
        )
    end
end

function ParaseleneWildRace.Options:AdjustVisibleWaypointsOptions()
    -- Check or uncheck the TrackMaster checkbox.
    if self.app.characterData.trackMasterOptions.enabled then
        self.btnTrackMaster:SetData(true) -- Checked. Used by mouse enter/exit handlers.
        self.wndTrackMasterIcon:SetSprite(
            (self.btnTrackMaster:ContainsMouse() and ParaseleneWildRace.checkedCheckboxFlybyIcon) or
            ParaseleneWildRace.checkedCheckboxIcon
        )
    else
        self.btnTrackMaster:SetData(false) -- Unchecked. Used by mouse enter/exit handlers.
        self.wndTrackMasterIcon:SetSprite(
            (self.btnTrackMaster:ContainsMouse() and ParaseleneWildRace.uncheckedCheckboxFlybyIcon) or
            ParaseleneWildRace.uncheckedCheckboxIcon
        )
    end
end

function ParaseleneWildRace.Options:HideWindow()
    if self.wndOptions:IsVisible() then
        self.wndOptions:Close()
        self.app.client.wndOptionsFrame:SetSprite(
            (self.app.client.btnOptions:ContainsMouse() and ParaseleneWildRace.enabledToolbarButtonFlybyFrame) or
            ParaseleneWildRace.enabledToolbarButtonFrame
        )
    end
end

function ParaseleneWildRace.Options:OnBtnAbout(wndHandler, wndControl, eMouseButton)
    self.app.about:ToggleWindowVisibility()
    Sound.Play(ParaseleneWildRace.buttonClickSound)
end

function ParaseleneWildRace.Options:OnBtnAboutGenerateTooltip(wndHandler, wndControl, eToolTipType, x, y)
    self.btnAbout:SetTooltip(
        (self.app.about.wndAbout:IsVisible() and "Close the \"About\" window") or
        "Open the \"About\" window"
    )
end

function ParaseleneWildRace.Options:OnBtnCircuit(wndHandler, wndControl, eMouseButton)
    if self.app.characterData.uiOptions.background ~= ParaseleneWildRace.backgrounds.circuit then
        self.app.characterData.uiOptions.background = ParaseleneWildRace.backgrounds.circuit
        self:SetBackgrounds()
        self:AdjustUserInterfaceOptions()
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    end
end

function ParaseleneWildRace.Options:OnBtnClose(wndHandler, wndControl, eMouseButton)
    self:HideWindow()
    Sound.Play(ParaseleneWildRace.buttonClickSound)
end

function ParaseleneWildRace.Options:OnBtnHolo(wndHandler, wndControl, eMouseButton)
    if self.app.characterData.uiOptions.background ~= ParaseleneWildRace.backgrounds.holo then
        self.app.characterData.uiOptions.background = ParaseleneWildRace.backgrounds.holo
        self:SetBackgrounds()
        self:AdjustUserInterfaceOptions()
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    end
end

function ParaseleneWildRace.Options:OnBtnLock(wndHandler, wndControl, eMouseButton)
    self:SetWindowLock(not self.app.characterData.optionsWindowInfo.locked)
    Sound.Play(ParaseleneWildRace.buttonClickSound)
end

function ParaseleneWildRace.Options:OnBtnLockGenerateTooltip(wndHandler, wndControl, eToolTipType, x, y)
    self.btnLock:SetTooltip(
        (self.app.characterData.optionsWindowInfo.locked and "Unlock the window so that it can be moved") or
        "Lock the window so that it cannot be moved"
    )
end

function ParaseleneWildRace.Options:OnBtnTrackMaster(wndHandler, wndControl, eMouseButton)
    self.app.trackMaster:ToggleEnable()
    Sound.Play(ParaseleneWildRace.buttonClickSound)
end

function ParaseleneWildRace.Options:OnWndTrackMasterGenerateTooltip(wndHandler, wndControl, eToolTipType, x, y)
    self.wndTrackMaster:SetTooltip(
        (self.app.characterData.trackMasterOptions.enabled and
         "Do not use TrackMaster to draw a line to the selected waypoint in the game world") or
         "Use TrackMaster to draw a line to the selected waypoint in the game world"
    )
end

function ParaseleneWildRace.Options:OnCheckboxMouseEnter(wndHandler, wndControl, x, y)
    local wndIcon = wndHandler:FindChild("wndIcon")
    local checked = wndHandler:GetData()
    if checked then
        wndIcon:SetSprite(ParaseleneWildRace.checkedCheckboxFlybyIcon)
    else
        wndIcon:SetSprite(ParaseleneWildRace.uncheckedCheckboxFlybyIcon)
    end
end

function ParaseleneWildRace.Options:OnCheckboxMouseExit(wndHandler, wndControl, x, y)
    local wndIcon = wndHandler:FindChild("wndIcon")
    local checked = wndHandler:GetData()
    if checked then
        wndIcon:SetSprite(ParaseleneWildRace.checkedCheckboxIcon)
    else
        wndIcon:SetSprite(ParaseleneWildRace.uncheckedCheckboxIcon)
    end
end

function ParaseleneWildRace.Options:OnRadioMouseEnter(wndHandler, wndControl, x, y)
    local wndIcon = wndHandler:FindChild("wndIcon")
    local checked = wndHandler:GetData()
    if checked then
        wndIcon:SetSprite(ParaseleneWildRace.checkedRadioFlybyIcon)
    else
        wndIcon:SetSprite(ParaseleneWildRace.uncheckedRadioFlybyIcon)
    end
end

function ParaseleneWildRace.Options:OnRadioMouseExit(wndHandler, wndControl, x, y)
    local wndIcon = wndHandler:FindChild("wndIcon")
    local checked = wndHandler:GetData()
    if checked then
        wndIcon:SetSprite(ParaseleneWildRace.checkedRadioIcon)
    else
        wndIcon:SetSprite(ParaseleneWildRace.uncheckedRadioIcon)
    end
end

function ParaseleneWildRace.Options:OnWindowMove(wndHandler, wndControl, nOldLeft, nOldTop, nOldRight, nOldBottom)
    -- Keep self.app.characterData.optionsWindowInfo up to date with the window's position.
    local left, top, right, bottom = self.wndOptions:GetRect()
    self.app.characterData.optionsWindowInfo.left = left or 0
    self.app.characterData.optionsWindowInfo.top = top or 0
end

function ParaseleneWildRace.Options:OnWndCircuitGenerateTooltip(wndHandler, wndControl, eToolTipType, x, y)
    self.wndCircuit:SetTooltip("Set the background of all the windows to the \"Circuit\" background")
end

function ParaseleneWildRace.Options:OnWndHoloGenerateTooltip(wndHandler, wndControl, eToolTipType, x, y)
    self.wndHolo:SetTooltip("Set the background of all the windows to the \"Holo\" background")
end

function ParaseleneWildRace.Options:SetBackgrounds()
    local sprite = ParaseleneWildRace.backgroundSprites[self.app.characterData.uiOptions.background]
    self.app.about.wndAbout:SetSprite(sprite)
    self.app.client.wndClient:SetSprite(sprite)
    self.app.editor.wndEditor:SetSprite(sprite)
    self.app.files.wndFiles:SetSprite(sprite)
    self.app.share.wndShare:SetSprite(sprite)
    self.wndOptions:SetSprite(sprite)
end

function ParaseleneWildRace.Options:SetWindowLock(isLocked)
    if isLocked then
        self.app.characterData.optionsWindowInfo.locked = true
        self.wndOptions:SetStyle("Moveable", false)
        self.wndLockIcon:SetSprite(ParaseleneWildRace.lockEnabledIcon)
    else
        self.app.characterData.optionsWindowInfo.locked = false
        self.wndOptions:SetStyle("Moveable", true)
        self.wndLockIcon:SetSprite(ParaseleneWildRace.lockDisabledIcon)
    end
end

function ParaseleneWildRace.Options:ShowWindow()
    if not self.wndOptions:IsVisible() then
        self.wndOptions:Invoke()
        self.app.client.wndOptionsFrame:SetSprite(
            (self.app.client.btnOptions:ContainsMouse() and ParaseleneWildRace.pressedToolbarButtonFlybyFrame) or
            ParaseleneWildRace.pressedToolbarButtonFrame
        )
    end
end

function ParaseleneWildRace.Options:ToggleWindowVisibility()
    if self.wndOptions:IsVisible() then
        self:HideWindow()
    else
        self:ShowWindow()
    end
end
