-- Paraselene's WildRace
-- Version 0.5.2

require "Apollo"
require "GameLib"
require "ICCommLib"
require "ICComm"
require "Window"

ParaseleneWildRace.Share = {}

function ParaseleneWildRace.Share:new(app)
    local i = {}
    setmetatable(i, self)
    self.__index = self

    i.app = app

    i.wndShare = Apollo.LoadForm(i.app.doc, "wndShare", nil, i)
    if not i.wndShare then
        local message = ParaseleneWildRace.appName ..
            ": Error: Could not load the \"Share\" window."
        ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_System, message)
        Apollo.AddAddonErrorText(ParaseleneWildRaceInstance, message)
        return nil
    end

    local wndTitle = i.wndShare:FindChild("wndTitle")
    i.btnAbout = wndTitle:FindChild("btnAbout")
    i.btnLock = wndTitle:FindChild("btnLock")
    i.wndLockIcon = i.btnLock:FindChild("wndIcon")

    local wndTools = i.wndShare:FindChild("wndTools")
    i.wndTitle = wndTools:FindChild("wndTitle")
    i.edtDetails = wndTools:FindChild("edtDetails")
    i.btnSend = wndTools:FindChild("btnSend")
    i.btnSend:SetData(ParaseleneWildRace.Button:new{
        button = i.btnSend
    })
    i.btnReceive = wndTools:FindChild("btnReceive")
    i.btnReceive:SetData(ParaseleneWildRace.Button:new{
        button = i.btnReceive,
        style = ParaseleneWildRace.buttonStyles.red
    })

    wndTitle:SetText(ParaseleneWildRace.appName .. " Share")

    local width = ParaseleneWildRace.shareWindowWidth
    local height = ParaseleneWildRace.shareWindowHeight
    i.wndShare:Move(
        i.app.characterData.shareWindowInfo.left,
        i.app.characterData.shareWindowInfo.top,
        width,
        height
    )
    i:SetWindowLock(i.app.characterData.shareWindowInfo.locked)

    self.state = self.InactiveStateEntry

    i.wndShare:SetSprite(ParaseleneWildRace.backgroundSprites[i.app.characterData.uiOptions.background])
    i.wndShare:Show(false, true)

    return i
end

function ParaseleneWildRace.Share:AlertStateEntry()
    self.wndTitle:SetText(self.title)
    self.edtDetails:SetText(self.details)

    self.inbound = nil
    self:DisableButtons()
    self.timeout = os.clock() + 5

    self.state = self.AlertStateLoop
    return self:AlertStateLoop()
end

function ParaseleneWildRace.Share:AlertStateLoop()
    if (not self.wndShare:IsVisible()) or (os.clock() >= self.timeout) then
        return self:InactiveStateEntry()
    end
end

function ParaseleneWildRace.Share:DisableButtons()
    self:DisableReceiveButton()
    self:DisableSendButton()
end

function ParaseleneWildRace.Share:DisableReceiveButton()
    self.btnReceive:GetData():Disable()
    self.btnReceive:Show(false, true)
    self.receivePressed = false
end

function ParaseleneWildRace.Share:DisableSendButton()
    self.btnSend:GetData():Disable()
    self.btnSend:Show(false, true)
    self.sendPressed = false
end

function ParaseleneWildRace.Share:EnableReceiveButton()
    self.btnReceive:GetData():Enable()
    self.btnReceive:Show(true, true)
    self.receivePressed = false
end

function ParaseleneWildRace.Share:EnableSendButton()
    self.btnSend:GetData():Enable()
    self.btnSend:Show(true, true)
    self.sendPressed = false
end

function ParaseleneWildRace.Share.GetTargetedPlayerInfo()
    local me = GameLib.GetPlayerUnit()
    local target = me and me:GetTarget()
    local name = target and target:GetName()
    local gender = target and target:GetGender()
    if target and target:IsACharacter() and name then
        return name, gender
    else
        return nil
    end
end

function ParaseleneWildRace.Share:HideWindow()
    if self.wndShare:IsVisible() then
        self.wndShare:Close()
        self.app.client.wndShareFrame:SetSprite(
            (self.app.client.btnShare:ContainsMouse() and ParaseleneWildRace.enabledToolbarButtonFlybyFrame) or
            ParaseleneWildRace.enabledToolbarButtonFrame
        )
    end
end

function ParaseleneWildRace.Share:InactiveStateEntry()
    self.title = ""
    self.details = ""
    self.wndTitle:SetText(self.title)
    self.edtDetails:SetText(self.details)

    self.inbound = nil
    self:DisableButtons()

    self.state = self.InactiveStateLoop
    return self:InactiveStateLoop()
end

function ParaseleneWildRace.Share:InactiveStateLoop()
    if self.wndShare:IsVisible() then
        return self:ReceiveReadyStateEntry()
    end
end

function ParaseleneWildRace.Share:OnBtnAbout(wndHandler, wndControl, eMouseButton)
    self.app.about:ToggleWindowVisibility()
    Sound.Play(ParaseleneWildRace.buttonClickSound)
end

function ParaseleneWildRace.Share:OnBtnAboutGenerateTooltip(wndHandler, wndControl, eToolTipType, x, y)
    self.btnAbout:SetTooltip(
        (self.app.about.wndAbout:IsVisible() and "Close the \"About\" window") or
        "Open the \"About\" window"
    )
end

function ParaseleneWildRace.Share:OnBtnClose(wndHandler, wndControl, eMouseButton)
    self:HideWindow()
    Sound.Play(ParaseleneWildRace.buttonClickSound)
end

function ParaseleneWildRace.Share:OnBtnLock(wndHandler, wndControl, eMouseButton)
    self:SetWindowLock(not self.app.characterData.shareWindowInfo.locked)
    Sound.Play(ParaseleneWildRace.buttonClickSound)
end

function ParaseleneWildRace.Share:OnBtnLockGenerateTooltip(wndHandler, wndControl, eToolTipType, x, y)
    self.btnLock:SetTooltip(
        (self.app.characterData.shareWindowInfo.locked and "Unlock the window so that it can be moved") or
        "Lock the window so that it cannot be moved"
    )
end

function ParaseleneWildRace.Share:OnBtnReceive(wndHandler, wndControl, eMouseButton)
    if self.btnReceive:GetData():IsEnabled() then
        self.receivePressed = true
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    end
end

function ParaseleneWildRace.Share:OnBtnSend(wndHandler, wndControl, eMouseButton)
    if self.btnSend:GetData():IsEnabled() then
        self.sendPressed = true
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    end
end

function ParaseleneWildRace.Share:OnButtonMouseEnter(wndHandler, wndControl, x, y)
    wndHandler:GetData():OnMouseEnter()
end

function ParaseleneWildRace.Share:OnButtonMouseExit(wndHandler, wndControl, x, y)
    wndHandler:GetData():OnMouseExit()
end

function ParaseleneWildRace.Share:OnCommMessageReceived(sender, message)
    if self.inbound then
        table.insert(self.inbound, {sender = sender, message = message})
    end
end

function ParaseleneWildRace.Share:OnUpdateTimer()
    self:state()
end

function ParaseleneWildRace.Share:OnWindowMove(wndHandler, wndControl, nOldLeft, nOldTop, nOldRight, nOldBottom)
    -- Keep self.app.characterData.shareWindowInfo up to date with the window's position.
    local left, top, right, bottom = self.wndShare:GetRect()
    self.app.characterData.shareWindowInfo.left = left or 0
    self.app.characterData.shareWindowInfo.top = top or 0
end

function ParaseleneWildRace.Share.ParseMessage(packet)
    local commType = string.sub(packet.message, 1, ParaseleneWildRace.commTypeLength)
    local id = string.sub(
        packet.message,
        ParaseleneWildRace.commTypeLength + 1,
        ParaseleneWildRace.commTypeLength + ParaseleneWildRace.commPacketIdDigits
    )
    local payload = string.sub(
        packet.message,
        ParaseleneWildRace.commTypeLength + ParaseleneWildRace.commPacketIdDigits + 1
    )
    return packet.sender, commType, id, payload
end

function ParaseleneWildRace.Share:ReceiveDataStateEntry()
    self:DisableReceiveButton()

    local message = ParaseleneWildRace.commShareType.accept .. self.id
    ParaseleneWildRace.Share.SendPrivateMessage(self.sender, message)
    self:SetTimeout(message)

    self.course = ParaseleneWildRace.Course:new{
        name = self.courseInfo.name,
        strictMode = self.courseInfo.strictMode,
        timedMode = self.courseInfo.timedMode,
        timeout = self.courseInfo.timeout
    }
    if self.courseInfo.waypointCount == 0 then
        self.app.client:SetCourse(self.course)
        self.details = "Transfer complete"
        Sound.Play(ParaseleneWildRace.commCompleteSound)
        return self:AlertStateEntry()
    end

    self.waypointIndex = 1

    self.SetDetails = function()
        self.details = "Receiving course\n\n" .. self.courseInfo.name .. "\n\nfrom\n" .. self.sender ..
            "\n\nWaiting for waypoint " .. self.waypointIndex .. " of " .. self.courseInfo.waypointCount ..
            "\n\nTimeout in " .. self:TimeRemaining() .. " seconds"
    end

    self.state = self.ReceiveDataStateLoop
    return self:ReceiveDataStateLoop()
end

function ParaseleneWildRace.Share:ReceiveDataStateLoop()
    self:SetDetails()
    self.edtDetails:SetText(self.details)

    if not self.wndShare:IsVisible() then
        return self:InactiveStateEntry()
    end
    if os.clock() >= self.timeout then
        self.details = "Transfer canceled\nNo response"
        Sound.Play(ParaseleneWildRace.commErrorSound)
        return self:AlertStateEntry()
    end
    while #self.inbound > 0 do
        local packet = self.inbound[1]
        table.remove(self.inbound, 1)
        local sender, commType, id, payload = ParaseleneWildRace.Share.ParseMessage(packet)
        if (sender == self.sender) and
           (commType == ParaseleneWildRace.commShareType.data) and (id == self.id) then
            local waypoint = ParaseleneWildRace.Deserialize.Waypoint(payload)
            if not waypoint then
                self.details = "Transfer canceled\nCommunication error"
                Sound.Play(ParaseleneWildRace.commErrorSound)
                return self:AlertStateEntry()
            end
            self.course:Append(waypoint)

            local message = ParaseleneWildRace.commShareType.dataOk .. self.id
            ParaseleneWildRace.Share.SendPrivateMessage(self.sender, message)
            self:SetTimeout(message)

            self.waypointIndex = self.waypointIndex + 1
            if self.waypointIndex <= self.courseInfo.waypointCount then
                return self:ReceiveDataStateLoop()
            else
                self.app.client:SetCourse(self.course)
                self.details = "Transfer complete"
                Sound.Play(ParaseleneWildRace.commCompleteSound)
                return self:AlertStateEntry()
            end
        end
    end
end

function ParaseleneWildRace.Share:ReceiveReadyStateEntry()
    self.title = "Receive a course"
    self.details = "Ready to receive a course"
    self.wndTitle:SetText(self.title)
    self.edtDetails:SetText(self.details)

    self.inbound = {}

    self.state = self.ReceiveReadyStateLoop
    return self:ReceiveReadyStateLoop()
end

function ParaseleneWildRace.Share:ReceiveReadyStateLoop()
    if not self.wndShare:IsVisible() then
        return self:InactiveStateEntry()
    end
    local targetName, targetGender = ParaseleneWildRace.Share.GetTargetedPlayerInfo()
    if targetName and self.app.client.course then
        self.targetName, self.targetGender = targetName, targetGender
        return self:SendReadyStateEntry()
    end
    while #self.inbound > 0 do
        local packet = self.inbound[1]
        table.remove(self.inbound, 1)
        local sender, commType, id, payload = ParaseleneWildRace.Share.ParseMessage(packet)
        if commType == ParaseleneWildRace.commShareType.request then
            self.sender, self.commType, self.id, self.payload = sender, commType, id, payload
            Sound.Play(ParaseleneWildRace.commRequestSound)
            return self:ReceiveRequestStateEntry()
        end
    end
end

function ParaseleneWildRace.Share:ReceiveRequestStateEntry()
    self.courseInfo = ParaseleneWildRace.Deserialize.CourseInfo(self.payload)
    if not self.courseInfo then
        self.details = "Transfer canceled\nCommunication error"
        Sound.Play(ParaseleneWildRace.commErrorSound)
        return self:AlertStateEntry()
    end

    self:DisableSendButton()
    self:EnableReceiveButton()

    self.title = "Receive a course"
    self.wndTitle:SetText(self.title)

    self.SetDetails = function()
        self.details = self.sender .. "\nwants to send you the course\n\n" ..
            self.courseInfo.name ..
            "\n\nWould you like to receive\nand load this course?\n\n" ..
            "Please answer within " .. self:TimeRemaining() .. " seconds"
    end

    self:SetTimeout(nil)

    self.state = self.ReceiveRequestStateLoop
    return self:ReceiveRequestStateLoop()
end

function ParaseleneWildRace.Share:ReceiveRequestStateLoop()
    self:SetDetails()
    self.edtDetails:SetText(self.details)

    if not self.wndShare:IsVisible() then
        return self:InactiveStateEntry()
    end
    if self.receivePressed then
        self.receivePressed = false
        return self:ReceiveDataStateEntry()
    end
    if os.clock() >= self.timeout then
        self.details = "Transfer canceled\nNo response"
        Sound.Play(ParaseleneWildRace.commErrorSound)
        return self:AlertStateEntry()
    end
end

function ParaseleneWildRace.Share:SendDataStateEntry()
    if #self.app.client.course.waypoints == 0 then
        self.details = "Transfer complete"
        Sound.Play(ParaseleneWildRace.commCompleteSound)
        return self:AlertStateEntry()
    end

    self.waypointIndex = 1

    local message = ParaseleneWildRace.commShareType.data .. self.id ..
        ParaseleneWildRace.Serialize.Waypoint(self.app.client.course.waypoints[self.waypointIndex])
    ParaseleneWildRace.Share.SendPrivateMessage(self.targetName, message)
    self:SetTimeout(message)

    self.SetDetails = function()
        self.details = "Sending the current course to\n" .. self.targetName ..
            "\n\nSending waypoint " .. self.waypointIndex .. " of " .. #self.app.client.course.waypoints .. "..." ..
            "\n\nTimeout in " .. self:TimeRemaining() .. " seconds"
    end

    self.state = self.SendDataStateLoop
    return self:SendDataStateLoop()
end

function ParaseleneWildRace.Share:SendDataStateLoop()
    self:SetDetails()
    self.edtDetails:SetText(self.details)

    if not self.wndShare:IsVisible() then
        return self:InactiveStateEntry()
    end
    if os.clock() >= self.timeout then
        self.details = "Transfer canceled\nNo response"
        Sound.Play(ParaseleneWildRace.commErrorSound)
        return self:AlertStateEntry()
    end
    while #self.inbound > 0 do
        local packet = self.inbound[1]
        table.remove(self.inbound, 1)
        local sender, commType, id, payload = ParaseleneWildRace.Share.ParseMessage(packet)
        if (sender == self.targetName) and
           (commType == ParaseleneWildRace.commShareType.dataOk) and (id == self.id) then
            self.waypointIndex = self.waypointIndex + 1
            if self.waypointIndex <= #self.app.client.course.waypoints then
                local message = ParaseleneWildRace.commShareType.data .. self.id ..
                    ParaseleneWildRace.Serialize.Waypoint(self.app.client.course.waypoints[self.waypointIndex])
                ParaseleneWildRace.Share.SendPrivateMessage(self.targetName, message)
                self:SetTimeout(message)
                return self:SendDataStateLoop()
            else
                self.details = "Transfer complete"
                Sound.Play(ParaseleneWildRace.commCompleteSound)
                return self:AlertStateEntry()
            end
        end
    end
end

function ParaseleneWildRace.Share.SendMessage(message)
    ParaseleneWildRaceInstance:SendMessage(ParaseleneWildRaceInstance.commType.share .. message)
end

function ParaseleneWildRace.Share.SendPrivateMessage(recipient, message)
    ParaseleneWildRaceInstance:SendPrivateMessage(recipient, ParaseleneWildRaceInstance.commType.share .. message)
end

function ParaseleneWildRace.Share:SendReadyStateEntry()
    self.title = "Receive or send a course"
    self.details = "Ready to receive a course\n\nor\n\nReady to send the current course to\n" .. self.targetName
    self.wndTitle:SetText(self.title)
    self.edtDetails:SetText(self.details)

    self:EnableSendButton()

    self.state = self.SendReadyStateLoop
    return self:SendReadyStateLoop()
end

function ParaseleneWildRace.Share:SendReadyStateLoop()
    if not self.wndShare:IsVisible() then
        return self:InactiveStateEntry()
    end
    local targetName, targetGender = ParaseleneWildRace.Share.GetTargetedPlayerInfo()
    if (not targetName) or (targetName ~= self.targetName) then
        return self:InactiveStateEntry()
    end
    if self.sendPressed then
        self.sendPressed = false
        return self:SendRequestStateEntry()
    end
    while #self.inbound > 0 do
        local packet = self.inbound[1]
        table.remove(self.inbound, 1)
        local sender, commType, id, payload = ParaseleneWildRace.Share.ParseMessage(packet)
        if commType == ParaseleneWildRace.commShareType.request then
            self.sender, self.commType, self.id, self.payload = sender, commType, id, payload
            Sound.Play(ParaseleneWildRace.commRequestSound)
            return self:ReceiveRequestStateEntry()
        end
    end
end

function ParaseleneWildRace.Share:SendRequestStateEntry()
    self:DisableSendButton()

    self.id = string.format("%03.0f", math.random(ParaseleneWildRace.commPacketIdMax))
    local message = ParaseleneWildRace.commShareType.request .. self.id ..
        ParaseleneWildRace.Serialize.CourseInfo(self.app.client.course)
    ParaseleneWildRace.Share.SendPrivateMessage(self.targetName, message)
    self:SetTimeout(message)

    self.title = "Send a course"
    self.wndTitle:SetText(self.title)

    self.SetDetails = function()
        self.details = "Asking\n" .. self.targetName .. "\nif " ..
            (((self.targetGender == Unit.CodeEnumGender.Male) and "he") or
             ((self.targetGender == Unit.CodeEnumGender.Female) and "she") or
             "it") .. " would like to receive\nthe current course...\n\nWaiting for a response from\n" ..
            self.targetName .. "\n\nTimeout in " .. self:TimeRemaining() .. " seconds"
    end

    self.state = self.SendRequestStateLoop
    return self:SendRequestStateLoop()
end

function ParaseleneWildRace.Share:SendRequestStateLoop()
    self:SetDetails()
    self.edtDetails:SetText(self.details)

    if not self.wndShare:IsVisible() then
        return self:InactiveStateEntry()
    end
    if os.clock() >= self.timeout then
        self.details = "Transfer canceled\nNo response"
        Sound.Play(ParaseleneWildRace.commErrorSound)
        return self:AlertStateEntry()
    end
    while #self.inbound > 0 do
        local packet = self.inbound[1]
        table.remove(self.inbound, 1)
        local sender, commType, id, payload = ParaseleneWildRace.Share.ParseMessage(packet)
        if (sender == self.targetName) and
           (commType == ParaseleneWildRace.commShareType.accept) and (id == self.id) then
            Sound.Play(ParaseleneWildRace.commAcceptSound)
            return self:SendDataStateEntry()
        end
    end
end

function ParaseleneWildRace.Share:SetTimeout(message)
    self.timeout = os.clock() + 30
    if message then
        local capacity = ICCommLib.GetUploadCapacityByType(ICCommLib.CodeEnumICCommChannelType.Global)
        if string.len(message) > capacity then
            if capacity < 0 then
                self.timeout = self.timeout + math.ceil(-capacity / 50)
            end
            self.timeout = self.timeout + math.ceil(string.len(message) / 50)
        end
    end
end

function ParaseleneWildRace.Share:SetWindowLock(isLocked)
    if isLocked then
        self.app.characterData.shareWindowInfo.locked = true
        self.wndShare:SetStyle("Moveable", false)
        self.wndLockIcon:SetSprite(ParaseleneWildRace.lockEnabledIcon)
    else
        self.app.characterData.shareWindowInfo.locked = false
        self.wndShare:SetStyle("Moveable", true)
        self.wndLockIcon:SetSprite(ParaseleneWildRace.lockDisabledIcon)
    end
end

function ParaseleneWildRace.Share:ShowWindow()
    if not self.wndShare:IsVisible() then
        self.wndShare:Invoke()
        self.app.client.wndShareFrame:SetSprite(
            (self.app.client.btnShare:ContainsMouse() and ParaseleneWildRace.pressedToolbarButtonFlybyFrame) or
            ParaseleneWildRace.pressedToolbarButtonFrame
        )
    end
end

function ParaseleneWildRace.Share:TimeRemaining()
    local now  = os.clock()
    return string.format("%.0f", ((self.timeout >= now) and (self.timeout - now)) or 0)
end

function ParaseleneWildRace.Share:ToggleWindowVisibility()
    if self.wndShare:IsVisible() then
        self:HideWindow()
    else
        self:ShowWindow()
    end
end
