-- Paraselene's WildRace
-- Version 0.5.2

ParaseleneWildRace.Storage = {}
 
function ParaseleneWildRace.Storage:new(o)
    if type(o) ~= "table" then
        return nil
    end

    local i = {}
    setmetatable(i, self)
    self.__index = self 

    i.baseFolder = ParaseleneWildRace.Folder:new(o.baseFolder)

    return (i:IsValid() and i) or nil
end

function ParaseleneWildRace.Storage:GetItem(path)
    return ParaseleneWildRace.Storage.GetItemInternal(self.baseFolder, path, 1)
end

function ParaseleneWildRace.Storage.GetItemInternal(folder, path, pathIdx)
    -- The folder UID better match the path UID.
    if folder.uid ~= path[pathIdx] then
        return nil
    end

    -- If this is the last UID in the path, then we are done.
    if pathIdx == #path then
        return folder
    end

    -- If the next UID in the path is the last UID in the path, then it could be a course.
    if (pathIdx + 1) == #path then
        for idx, course in ipairs(folder.courses) do
            if course.uid == path[pathIdx + 1] then
                return course
            end
        end
        -- It was not a course. It could be a leaf folder. Fall through.
    end

    -- The next UID in the path must be a folder.
    for idx, childFolder in ipairs(folder.folders) do
        if childFolder.uid == path[pathIdx + 1] then
            return ParaseleneWildRace.Storage.GetItemInternal(childFolder, path, pathIdx + 1)
        end
    end

    return nil -- There is no item at the given path.
end

--[[
function ParaseleneWildRace.Storage.CreatePathString(path)
    local pathString = ""
    for idx, name in ipairs(path) do
        if idx > 1 then
            pathString = pathString .. "/"
        end
        pathString = pathString .. name
    end
    return "{" .. pathString .. "}"
end
--]]

function ParaseleneWildRace.Storage.GetPath(item)
    if item.parent then
        local newPath = ParaseleneWildRace.Storage.GetPath(item.parent)
        table.insert(newPath, item.uid)
        return newPath
    else
        return {item.uid}
    end
end

function ParaseleneWildRace.Storage:GetFamily(path)
    local item = self:GetItem(path)
    local parent = item and item.parent
    local siblings
    if parent then
        if item.folderItemType == ParaseleneWildRace.folderItemTypes.folder then
            siblings = parent.folders
        else -- item.folderItemType == ParaseleneWildRace.folderItemTypes.course
            siblings = parent.courses
        end
    end
    local index
    if siblings then
        for idx, childItem in ipairs(siblings) do
            if childItem.uid == item.uid then
                index = idx
                break
            end
        end
    end
    return item, parent, siblings, index
end

function ParaseleneWildRace.Storage:IsValid()
    return (self.baseFolder ~= nil)
end

function ParaseleneWildRace.Storage:Save()
    return {
        baseFolder = self.baseFolder:Save()
    }
end

function ParaseleneWildRace.Storage:ShiftEarlier(path)
    local _, _, siblings, index = self:GetFamily(path)
    siblings[index - 1], siblings[index] = siblings[index], siblings[index - 1]
end

function ParaseleneWildRace.Storage:ShiftLater(path)
    local _, _, siblings, index = self:GetFamily(path)
    siblings[index], siblings[index + 1] = siblings[index + 1], siblings[index]
end
