-- Paraselene's WildRace
-- Version 0.5.2

require "Apollo"

ParaseleneWildRace.TrackMaster = {}

function ParaseleneWildRace.TrackMaster:new(app)
    local i = {}
    setmetatable(i, self)
    self.__index = self

    i.app = app

    return i
end

function ParaseleneWildRace.TrackMaster:Connect()
    local connectionJustEstablished = false
    if not self.addon then
        self.addon = Apollo.GetAddon("TrackMaster")
        if self.addon then
            connectionJustEstablished = true
            self.addonConfig = self.addon:AddToConfigMenu(
                self.addon.Type.Track,
                ParaseleneWildRace.appName,
                {
                    CanFire = false,
                    CanEnable = true,
                    IsChecked = self.app.characterData.trackMasterOptions.enabled,
                    OnEnableChanged = (
                        function(enable)
                            if (enable ~= self.app.characterData.trackMasterOptions.enabled) then
                                self:SetEnable(enable)
                            end
                        end
                    ),
                    LineNo = self.app.characterData.trackMasterOptions.lineNumber,
                    OnLineChanged = (
                        function(lineNumber)
                            if (lineNumber ~= self.app.characterData.trackMasterOptions.lineNumber) then
                                self:SetLineNumber(lineNumber)
                            end
                        end
                    )
                }
            )
        end
    end
    return connectionJustEstablished
end

function ParaseleneWildRace.TrackMaster:SetEnable(enable)
    if (enable ~= self.app.characterData.trackMasterOptions.enabled) then
        if self.app.characterData.trackMasterOptions.enabled and self.target and self.addon then
            self.addon:SetTarget(nil, -1, self.app.characterData.trackMasterOptions.lineNumber)
            self.target = nil
        end
        self.app.characterData.trackMasterOptions.enabled = enable
        if self.addon then
            self.addonConfig:SetEnabled(self.app.characterData.trackMasterOptions.enabled)
        end
        self.app.options:AdjustVisibleWaypointsOptions()
    end
end

function ParaseleneWildRace.TrackMaster:SetLineNumber(lineNumber)
    if (lineNumber ~= self.app.characterData.trackMasterOptions.lineNumber) then
        if self.app.characterData.trackMasterOptions.enabled and self.target and self.addon then
            self.addon:SetTarget(nil, -1, self.app.characterData.trackMasterOptions.lineNumber)
        end
        self.app.characterData.trackMasterOptions.lineNumber = lineNumber
        if self.app.characterData.trackMasterOptions.enabled and self.target and self.addon then
            self.addon:SetTarget(self.target, -1, self.app.characterData.trackMasterOptions.lineNumber)
        end
    end
end

function ParaseleneWildRace.TrackMaster:SetTarget(target, currentPosition)
    if self.app.characterData.trackMasterOptions.enabled and self.addon then
        if target then
            if currentPosition and
               target.isVisitable and 
               (target.zone == currentPosition.zone) and
               (target.inHousing == currentPosition.inHousing) then
                if not ParaseleneWildRace.Waypoint.IsEqual(target, self.target) then
                    self.addon:SetTarget(nil, -1, self.app.characterData.trackMasterOptions.lineNumber)
                    local vector = Vector3.New(target.point.x, target.point.y, target.point.z)
                    self.addon:SetTarget(vector, -1, self.app.characterData.trackMasterOptions.lineNumber)
                    self.target = target
                end
            elseif self.target then
                self.addon:SetTarget(nil, -1, self.app.characterData.trackMasterOptions.lineNumber)
                self.target = nil
            end
        elseif self.target then
            self.addon:SetTarget(nil, -1, self.app.characterData.trackMasterOptions.lineNumber)
            self.target = nil
        end
    end
end

function ParaseleneWildRace.TrackMaster:ToggleEnable()
    self:SetEnable(not self.app.characterData.trackMasterOptions.enabled)
end
