-- Paraselene's WildRace
-- Version 0.5.2

require "GameLib"
require "HousingLib"

ParaseleneWildRace.Waypoint = {}
 
function ParaseleneWildRace.Waypoint:new(o)
    if type(o) ~= "table" then
        return nil
    end

    local i = {}
    setmetatable(i, self)
    self.__index = self 

    i.name = ((type(o.name) == "string") and o.name) or nil
    i.realm = ((type(o.realm) == "string") and o.realm) or nil
    i.zone = ((type(o.zone) == "string") and o.zone) or nil
    i.subZone = ((type(o.zone) == "string") and o.subZone) or nil

    if type(o.inHousing) == "boolean" then
        i.inHousing = o.inHousing
    end

    if (type(o.faction) == "number") and
       ((o.faction == ParaseleneWildRace.waypointFaction.all) or
        (o.faction == ParaseleneWildRace.waypointFaction.dominion) or
        (o.faction == ParaseleneWildRace.waypointFaction.exile)) then
        i.faction = o.faction
    end

    i.point = ParaseleneWildRace.Point:new(o.point)

    return (i:IsValid() and i) or nil
end

function ParaseleneWildRace.Waypoint.GetCurrentPosition()
    local zoneInfo = GameLib.GetCurrentZoneMap()

    local realm = ParaseleneWildRaceInstance.realm
    local zone = zoneInfo and zoneInfo.strName
    zone = zone and string.gsub(zone, "^%s*(.-)%s*$", "%1") -- Strip leading and trailing spaces.
    local subZone = GetCurrentSubZoneName()
    subZone = subZone and string.gsub(subZone, "^%s*(.-)%s*$", "%1") -- Strip leading and trailing spaces.
    local inHousing = HousingLib.IsHousingWorld()
    local faction = (inHousing and ParaseleneWildRaceInstance.faction) or ParaseleneWildRace.waypointFaction.all
    local point = ParaseleneWildRace.Point.GetCurrentPosition()

    if inHousing then
        if HousingLib.IsOnMyResidence() then
            zone = ParaseleneWildRaceInstance.name
        else
            zone = subZone and subZone:match("%[(.*)%]")
            subZone = subZone and subZone:match("%s*(.-)%s*%[.*%]")
        end
    end
    local name = subZone

    return ParaseleneWildRace.Waypoint:new{
        name = name,
        realm = realm,
        zone = zone,
        subZone = subZone,
        inHousing = inHousing,
        faction = faction,
        point = point
    }
end

function ParaseleneWildRace.Waypoint.IsEqual(waypoint1, waypoint2)
    return (waypoint1 == waypoint2) or
           (waypoint1 and waypoint2 and
            (waypoint1.name == waypoint2.name) and
            (waypoint1.realm == waypoint2.realm) and
            (waypoint1.zone == waypoint2.zone) and
            (waypoint1.subZone == waypoint2.subZone) and
            (waypoint1.inHousing == waypoint2.inHousing) and
            (waypoint1.faction == waypoint2.faction) and
            ParaseleneWildRace.Point.IsEqual(waypoint1.point, waypoint2.point))
end

function ParaseleneWildRace.Waypoint:IsValid()
    return (self.name ~= nil) and
           (self.realm ~= nil) and
           (self.zone ~= nil) and
           (self.subZone ~= nil) and
           (self.inHousing ~= nil) and
           (self.faction ~= nil) and
           (self.point ~= nil)
end

function ParaseleneWildRace.Waypoint:IsVisitable()
    if self.inHousing then
        return (self.realm == ParaseleneWildRaceInstance.realm) --[[ and
               (self.faction == ParaseleneWildRaceInstance.faction) --]]
    elseif (ParaseleneWildRaceInstance.faction == ParaseleneWildRace.waypointFaction.dominion) then
        if ((self.zone == "Exile-only zone 1") and (self.subZone == "Exile-only sub-zone 1")) or -- combination.
           (self.zone == "Exile-only zone 2") or
           (self.subZone == "Exile-only sub-zone 3") then
            return false
        end
    elseif (ParaseleneWildRaceInstance.faction == ParaseleneWildRace.waypointFaction.exile) then
        if ((self.zone == "Dominion-only zone 1") and (self.subZone == "Dominion-only sub-zone 1")) or -- combination.
           (self.zone == "Dominion-only zone 2") or
           (self.subZone == "Dominion-only sub-zone 3") then
            return false
        end
    end
    return true
end

function ParaseleneWildRace.Waypoint:Save()
    return {
        name = self.name,
        realm = self.realm,
        zone = self.zone,
        subZone = self.subZone,
        inHousing = self.inHousing,
        faction = self.faction,
        point = self.point:Save()
    }
end
